from . import Decimal

Amount = int

ONE = 2 ** 32

def encode(x): return x.sqrt() * ONE
def decode(x): return (x / ONE) ** 2

def mulDivF(x, y, z): return x * y // z
def mulDivC(x, y, z): return (x * y + z - 1) // z

class Order:
    def __init__(self, order):
        liq = int(Decimal(order['liquidity']))
        min = int(encode(Decimal(order['lowestRate'])))
        max = int(encode(Decimal(order['highestRate'])))
        mid = int(encode(Decimal(order['marginalRate'])))
        self.y = liq
        self.z = liq * (max - min) // (mid - min)
        self.A = max - min
        self.B = min
    def __iter__(self):
        y = Decimal(self.y)
        z = Decimal(self.z)
        A = Decimal(self.A)
        B = Decimal(self.B)
        yield 'liquidity'    , y
        yield 'lowestRate'   , decode(B)
        yield 'highestRate'  , decode(B + A)
        yield 'marginalRate' , decode(B + A * y / z)

def tradeBySourceAmount(x, order):
    y, z, A, B = [order.y, order.z, order.A, order.B]
    temp1 = y * A + z * B
    temp2 = temp1 * x // ONE
    temp3 = temp2 * A + z * z * ONE
    return x, mulDivF(temp1, temp2, temp3)

def tradeByTargetAmount(x, order):
    y, z, A, B = [order.y, order.z, order.A, order.B]
    temp1 = z * ONE
    temp2 = y * A + z * B
    temp3 = temp2 - x * A
    return mulDivC(x * temp1, temp1, temp2 * temp3), x



# ------------------------------------------------------------
# Auto generated test file `test_035_UsePositions.py`
# ------------------------------------------------------------
# source file   = NBTest_035_UsePositions.py
# source path   = /Users/skl/REPOES/Bancor/CarbonSimulator/resources/NBTest/
# target path   = /Users/skl/REPOES/Bancor/CarbonSimulator/resources/NBTest/
# test id       = 035
# test comment  = UsePositions
# ------------------------------------------------------------



from carbon import CarbonSimulatorUI, P, __version__, __date__
print(f"Carbon v{__version__} ({__date__})")
print("{0.__name__} v{0.__VERSION__} ({0.__DATE__})".format(CarbonSimulatorUI))





# ------------------------------------------------------------
# Test      035
# File      test_035_UsePositions.py
# Segment   use_position
# ------------------------------------------------------------
def test_use_position():
# ------------------------------------------------------------
    
    Sim = CarbonSimulatorUI(verbose=False, raiseonerror=False, pair="ETH/USDC")
    Sim
    
    for i in range(10):
        Sim.add_order("ETH", 10, 2000+i, 2200)
    
    r = Sim.state()["orders"].query("disabled==False")
    assert len(r) == 10
    assert sum(r.query("tkn=='ETH'")["y"]) == 100
    r
    
    help(Sim.amm_sells)
    
    r = Sim.amm_sells("ETH", 10, execute=False, pair="ETH/USDC")["trades"]
    p = r.iloc[-1]["price"]
    print(p)
    r
    
    r = Sim.amm_sells("ETH", 10, execute=False, use_positions=[2])["trades"]
    p1 = r.iloc[-1]["price"]
    print(p1, p)
    assert p1>p
    assert r.iloc[-1]["routeix"] == "[2]"
    r
    
    r = Sim.amm_sells("ETH", 10, execute=False, use_positions=[4])["trades"]
    p2 = r.iloc[-1]["price"]
    print(p2, p1, p)
    assert p2>p1
    assert r.iloc[-1]["routeix"] == "[4]"
    r
    
    r = Sim.amm_sells("ETH", 10, execute=False, use_positions=[6,8])["trades"]
    p3 = r.iloc[-1]["price"]
    print(p3, p2, p1, p)
    assert p3>p
    assert r.iloc[-1]["routeix"] == "[6, 8]"
    r
    
    r = Sim.amm_sells("ETH", 10, execute=False, use_positions=[12,14])["trades"]
    p4 = r.iloc[-1]["price"]
    print(p4, p3, p2, p1, p)
    assert p4>p3
    assert r.iloc[-1]["routeix"] == "[12, 14]"
    r
    
    r.iloc[-1]["routeix"]
    
    


# ------------------------------------------------------------
# Auto generated test file `test_044_MarginalPriceRouting.py`
# ------------------------------------------------------------
# source file   = NBTest_044_MarginalPriceRouting.py
# source path   = /Users/skl/REPOES/Bancor/CarbonSimulator/resources/NBTest/
# target path   = /Users/skl/REPOES/Bancor/CarbonSimulator/resources/NBTest/
# test id       = 044
# test comment  = MarginalPriceRouting
# ------------------------------------------------------------



from carbon import CarbonSimulatorUI, CarbonOrderUI, P, __version__, __date__
from math import sqrt
import numpy as np
from matplotlib import pyplot as plt
print(f"Carbon v{__version__} ({__date__})")
print("{0.__name__} v{0.__VERSION__} ({0.__DATE__})".format(CarbonSimulatorUI))
print("{0.__name__} v{0.__VERSION__} ({0.__DATE__})".format(CarbonOrderUI))






# ------------------------------------------------------------
# Test      044
# File      test_044_MarginalPriceRouting.py
# Segment   CarbonPair related tests
# ------------------------------------------------------------
def test_carbonpair_related_tests():
# ------------------------------------------------------------
    
    pair = P("ETH/USDC")
    assert isinstance(pair, P)
    print(str(pair))
    assert str(pair) == "P('ETH/USDC')"
    pair2 = P.from_slashpair("  ETH / USDC")
    print(pair2)
    assert isinstance(pair2, P)
    assert pair2.slashpair == "ETH/USDC"
    pair3 = P(pair2)
    assert isinstance(pair3, P)
    assert pair3==pair2
    

# ------------------------------------------------------------
# Test      044
# File      test_044_MarginalPriceRouting.py
# Segment   CarbonOrderUI general tests
# ------------------------------------------------------------
def test_carbonorderui_general_tests():
# ------------------------------------------------------------
    
    order = CarbonOrderUI.from_prices("ETH/USDC", "ETH", 2000, 3000, 10, 10)
    assert isinstance(order, CarbonOrderUI)
    #print(round(order.B,10))
    assert round(order.B,10) == 0.0182574186
    #print(round(order.S,10))
    assert round(order.S,10) == 0.0041032612
    assert order.yint == 10
    assert order.y == 10
    assert order.pa == 2000
    assert order.pb == 3000
    assert order.p_marg == 2000
    assert order.pa == order.py
    assert order.pb == order.px
    assert order.total_liquidity == (10, 'ETH')
    assert round(order.p0, 10) == round(sqrt(2000*3000),10)
    assert round(order.widthpc, 10) == round((3000-2000)/sqrt(2000*3000),10)
    assert round(order.widthr, 10) == round(3000/2000,10)
    order
    
    try:
        CarbonOrderUI.from_prices("ETH/USDC", "ETH", 2000, 3000, 5, 10)
        raise RuntimeError("Should have raised exception")
    except ValueError as e:
        print(e)
    
    try:
        CarbonOrderUI.from_prices("ETH/USDC", "ETH", 2000, 3000, -10, -10)
        raise RuntimeError("Should have raised exception")
    except ValueError as e:
        print(e)
    
    try:
        CarbonOrderUI.from_prices("ETH/USDC", "ETH", 2000, 3000, 10, -10)
        raise RuntimeError("Should have raised exception")
    except ValueError as e:
        print(e)
    
    order = CarbonOrderUI.from_prices("ETH/USDC", "ETH", 3000, 2000, 10, 10)
    assert order.pa == 2000
    assert order.pb == 3000
    order
    
    orderr = CarbonOrderUI.from_prices("ETH/USDC", "USDC", 1000, 750, 10000, 10000)
    assert orderr.pa == 1000
    assert round(orderr.pb,6) == 750
    orderr
    
    assert order.p_marg_f(0) == order.p_marg
    assert round(order.p_marg_f(5), 6) == 2424.492346
    assert order.p_marg_f(10) == 3000
    
    assert orderr.p_marg_f(0) == orderr.p_marg
    assert round(orderr.p_marg_f(10000),8) == 750
    
    assert order.p_marg_f(-1, raiseonerror=False) is None
    assert order.p_marg_f(10.0001, raiseonerror=False) is None
    assert order.p_marg_f(-1) is None
    assert order.p_marg_f(10.0001) is None
    
    assert orderr.p_marg_f(10001, raiseonerror=False) is None
    assert orderr.p_marg_f(10001) is None
    
    try:
        order.p_marg_f(-1, raiseonerror=True)
        raise RuntimeError("Should have raised exception")
    except ValueError as e:
        print(e)
        assert str(e) == "('Trade size dy must be a non-negative number', -1)"
    
    try:
        order.p_marg_f(10.001, raiseonerror=True)
        raise RuntimeError("Should have raised exception")
    except ValueError as e:
        print(e)
    
    try:
        order.p_marg_f(10001, raiseonerror=True)
        raise RuntimeError("Should have raised exception")
    except ValueError as e:
        print(e)
    

# ------------------------------------------------------------
# Test      044
# File      test_044_MarginalPriceRouting.py
# Segment   CarbonOrderUI tests yfromp
# ------------------------------------------------------------
def test_carbonorderui_tests_yfromp():
# ------------------------------------------------------------
    
    order = CarbonOrderUI.from_prices("ETH/USDC", "ETH", 2000, 3000, 10, 10)
    order1 = CarbonOrderUI.from_prices("ETH/USDC", "ETH", 2000, 3000, 10, 5)
    orderr = CarbonOrderUI.from_prices("ETH/USDC", "USDC", 1000, 750, 10000, 10000)
    
    assert order.yfromp_f(2000) == 10
    assert order.yfromp_f(3000) == 0
    assert order.dyfromp_f(2000) == 0
    assert order.dyfromp_f(3000) == 10
    
    try:
        order1.yfromp_f(2000, raiseonerror=True)
        raise RuntimeError("Should have raised exception")
    except ValueError as e:
        print(e)
    
    assert orderr.yfromp_f(1000) == 10000
    assert orderr.yfromp_f(750.00000001) < 1e-6
    assert orderr.dyfromp_f(1000) == 0
    assert round(orderr.dyfromp_f(750.0000001),5)==10000
    
    for dy in np.linspace(0,10):
        yy = order.yfromp_f(order.p_marg_f(dy))
        assert round(yy - (10-dy),10) == 0
        #print (f"dy={dy}, y={yy}, 10-dy={10-dy}")
    
    for dy in np.linspace(0,10):
        dy1 = order.dyfromp_f(order.p_marg_f(dy))
        assert round(dy - dy1,10) == 0
        #print (f"dy={dy}, dy1={dy1}")
    
    for dy in np.linspace(0,10000):
        dy1 = orderr.dyfromp_f(orderr.p_marg_f(dy))
        assert round(dy - dy1,4) == 0
        #print (f"dy={dy}, dy1={dy1}")
    

# ------------------------------------------------------------
# Test      044
# File      test_044_MarginalPriceRouting.py
# Segment   CarbonOrderUI tests dyfromdx_f and dxfromdy_f
# ------------------------------------------------------------
def test_carbonorderui_tests_dyfromdx_f_and_dxfromdy_f():
# ------------------------------------------------------------
    
    help(CarbonOrderUI.dyfromdx_f)
    
    # +
    # in order, order1 dx is a USDC number and dy is an ETH number
    order = CarbonOrderUI.from_prices("ETH/USDC", "ETH", 2000, 3000, 10, 10)
    order1 = CarbonOrderUI.from_prices("ETH/USDC", "ETH", 2000, 3000, 10, 5)
    
    # in orderr dx is an ETH number and dy is a USDC number
    orderr = CarbonOrderUI.from_prices("ETH/USDC", "USDC", 1000, 750, 10000, 10000)
    dxeps = dyeps = 0.000001
    p0 = sqrt(2000*3000)
    p0r = sqrt(1000*750)
    dxmax = 10*p0-dxeps
    dxmaxr = 10000/p0r-dxeps
    p0, p0r, dxmax, dxmaxr
    # -
    
    # ### dyfromdx_f
    
    dy = order.dyfromdx_f(dxeps)
    assert round(dxeps/dy,6) == 2000
    dy/dxeps, dxeps/dy
    
    dy = order.dyfromdx_f(dxmax)
    assert round(dxmax/dy - p0,6) == 0
    dy/dxmax, dxmax/dy, p0
    
    dy = order1.dyfromdx_f(dxeps)
    assert dxeps/dy > 2000
    assert dxeps/dy < 3000
    dy/dxeps, dxeps/dy
    
    dy = orderr.dyfromdx_f(dxeps)
    assert round(dy/dxeps,4) == 1000
    dy/dxeps, dxeps/dy
    
    dy = orderr.dyfromdx_f(dxmaxr)
    assert round(dy/dxmaxr - p0r,4) == 0
    dy/dxmaxr, dxmaxr/dy, p0r
    
    # ### dxfromdy_f
    
    dx = order.dxfromdy_f(dyeps)
    assert round(dx/dyeps,4) == 2000
    dx/dyeps, dyeps/dx
    
    dx = order.dxfromdy_f(10)
    assert round(dx/10 - p0,6) == 0
    dx/10, 10/dx, p0
    
    dx = order1.dxfromdy_f(dyeps)
    assert dx/dyeps > 2000
    assert dx/dyeps < 3000
    dx/dyeps, dyeps/dx
    
    dx = orderr.dxfromdy_f(dyeps)
    assert round(dyeps/dx,4) == 1000
    dx/dyeps, dyeps/dx
    
    dx = orderr.dxfromdy_f(10000)
    assert round(10000/dx - p0r,4) == 0
    dx/10000, 10000/dx, p0r
    
    # ### xfromy_f
    
    assert order.xfromy_f(order.yint) == 0
    assert round(order.xfromy_f(0)- order.p0*order.yint, 10) == 0
    assert order.x == 0
    assert order.xfromy_f(order.y) == order.x
    for i in range(10):
        #print (i)
        assert round(order.dxfromdy_f(i) - order.xfromy_f(order.y-i), 10) == 0
    
    assert orderr.xfromy_f(orderr.yint) == 0
    assert round(orderr.xfromy_f(0)- orderr.yint/orderr.p0, 10) == 0
    assert orderr.xfromy_f(orderr.y) == orderr.x
    for i in range(10):
        #print (i)
        assert round(orderr.dxfromdy_f(i*1000) - orderr.xfromy_f(orderr.y-i*1000), 10) == 0
    
    assert order1.xfromy_f(order1.yint) == 0
    assert round(order1.xfromy_f(0)- order.p0*order.yint, 10) == 0
    assert order1.xfromy_f(order1.y) == order1.x
    for i in range(5):
        #print (i)
        assert round(order1.xfromy_f(order1.y) + order1.dxfromdy_f(i) - order1.xfromy_f(order1.y-i), 10) == 0
    
    # ### yfromx_f
    
    assert order.yfromx_f(0) == order.yint
    assert round(order.yfromx_f(order.xint),5) == 0
    for i in range(10):
        #print(i)
        assert round(order.yfromx_f(order.xfromy_f(i)) - i, 10)  == 0
    
    order.yfromx_f(order.xint)
    
    
    
    assert orderr.yfromx_f(0) == orderr.yint
    assert round(orderr.yfromx_f(orderr.xint-0.00000001),4)==0
    for i in range(1,10):
        #print(i)
        assert round(orderr.yfromx_f(orderr.xfromy_f(i*1000)) - i*1000, 10)  == 0
    
    # ### p_eff_f
    
    assert order.p_eff_f(0) == 2000
    assert round(order.p_eff_f(0.000000001)-2000,3) == 0
    assert round(order.p_eff_f(10) - order.p0, 6) == 0
    
    p1 = order1.p_eff_f(0)
    assert p1 == order1.p_marg_f(0)
    p2 = order1.p_marg_f(5)
    assert round(order1.p_eff_f(5) - sqrt(p1*p2),5) == 0
    p1,p2, sqrt(p1*p2)
    
    assert orderr.p_eff_f(0) == 1000
    assert round(orderr.p_eff_f(0.1)-1000,2) == 0
    assert round(orderr.p_eff_f(10000) - orderr.p0, 6) == 0
    
    # ### from_Qxy 
    
    assert order.Q == sqrt(order.pb_raw/order.pa_raw)
    assert order.Gamma == 1 - sqrt(order.Q)
    print(f"Q={order.Q}, Gamma={order.Gamma}")
    
    assert order1.Q == sqrt(order1.pb_raw/order1.pa_raw)
    assert order1.Gamma == 1 - sqrt(order1.Q)
    print(f"Q={order1.Q}, Gamma={order1.Gamma}")
    
    assert orderr.Q == sqrt(orderr.pb_raw/orderr.pa_raw)
    assert orderr.Gamma == 1 - sqrt(orderr.Q)
    print(f"Q={orderr.Q}, Gamma={orderr.Gamma}")
    
    order_ = CarbonOrderUI.from_Qxy(order.pair, order.tkn, order.Q, order.xint, order.yint, order.y)
    assert round(order_.B - order.B, 10) == 0
    assert round(order_.S - order.S, 10) == 0
    assert round(order_.xint - order.xint, 10) == 0
    assert round(order_.yint - order.yint, 10) == 0
    assert round(order_.y - order.y, 10) == 0
    order_
    
    orderr_ = CarbonOrderUI.from_Qxy(orderr.pair, orderr.tkn, orderr.Q, orderr.xint, orderr.yint, orderr.y)
    assert round(orderr_.B - orderr.B, 10) == 0
    assert round(orderr_.S - orderr.S, 10) == 0
    assert round(orderr_.xint - orderr.xint, 10) == 0
    assert round(orderr_.yint - orderr.yint, 10) == 0
    assert round(orderr_.y - orderr.y, 10) == 0
    orderr_
    
    # ### buyx
    
    order_ = CarbonOrderUI.from_order(order)
    r = order_.buyx(1000, raiseonerror=True)
    r
    
    assert round(r["x"],10) == 1000
    assert round(r["p"] - sqrt(r["pmarg_old"]*r["pmarg"]),10) == 0
    assert round(r["dx/dy"] * r["dy/dx"],10) == 1
    assert r["tx"] == 'Sell 0.4954541237152398 ETH buy USDC'
    assert r["expanded"] == False
    assert r["y_old"] - r["dy"] == r["y"]
    assert round(r["x"] - r["dx"],10)==0
    
    r  = order_.buyx(1000, raiseonerror=True)
    r
    
    assert round(r["x"],10) == 2000
    assert round(r["p"] - sqrt(r["pmarg_old"]*r["pmarg"]),10) == 0
    assert r["dx/dy"] == 1 / r["dy/dx"]
    assert r["tx"] == 'Sell 0.4865262015696128 ETH buy USDC'
    assert r["expanded"] == False
    assert r["y_old"] - r["dy"] == r["y"]
    assert round(r["x"] - r["dx"],10)!=0
    
    order_ = CarbonOrderUI.from_order(orderr)
    r  = order_.buyx(1, raiseonerror=True)
    r
    
    assert round(r["x"],10) == 1
    assert round(r["p"] - sqrt(r["pmarg_old"]*r["pmarg"]),10) == 0
    assert r["dx/dy"] == 1 / r["dy/dx"]
    assert r["tx"] == 'Sell 986.7796593583731 USDC buy ETH'
    assert r["expanded"] == False
    assert r["y_old"] - r["dy"] == r["y"]
    assert round(r["x"] - r["dx"],10)==0
    
    r  = order_.buyx(-0.5, raiseonerror=True)
    r
    
    assert round(r["x"],10) == 0.5
    assert round(r["p"] - sqrt(r["pmarg_old"]*r["pmarg"]),10) == 0
    assert r["dx/dy"] == 1 / r["dy/dx"]
    assert r["tx"] == 'Buy 490.10673706911854 USDC sell ETH'
    assert r["expanded"] == False
    assert r["y_old"] - r["dy"] == r["y"]
    assert round(r["x"] - r["dx"],10)!=0
    
    # ### selly
    
    order_ = CarbonOrderUI.from_order(order)
    r  = order_.selly(1, raiseonerror=True)
    r
    
    assert r["y"] == 9
    assert round(r["p"] - sqrt(r["pmarg_old"]*r["pmarg"]),10) == 0
    assert r["dx/dy"] == 1 / r["dy/dx"]
    assert r["tx"] == 'Sell 1 ETH buy USDC'
    assert r["expanded"] == False
    assert r["y_old"] - r["dy"] == r["y"]
    assert round(r["x"] - r["dx"],10)==0
    
    r  = order_.selly(1, raiseonerror=True)
    r
    
    assert r["y"] == 8
    assert round(r["p"] - sqrt(r["pmarg_old"]*r["pmarg"]),10) == 0
    assert r["dx/dy"] == 1 / r["dy/dx"]
    assert r["tx"] == 'Sell 1 ETH buy USDC'
    assert r["expanded"] == False
    assert r["y_old"] - r["dy"] == r["y"]
    assert round(r["x"] - r["dx"],10)!=0
    
    order_ = CarbonOrderUI.from_order(order)
    r1 = order.selly(1, execute = False)
    r2 = order.selly(1, execute = False)
    assert r1 == r2
    
    order_ = CarbonOrderUI.from_order(orderr)
    r  = order_.selly(1000, raiseonerror=True)
    r
    
    assert r["y"] == 9000
    assert round(r["p"] - sqrt(r["pmarg_old"]*r["pmarg"]),10) == 0
    assert r["dx/dy"] == 1 / r["dy/dx"]
    assert r["tx"] == 'Sell 1000 USDC buy ETH'
    assert r["expanded"] == False
    assert r["y_old"] - r["dy"] == r["y"]
    assert round(r["x"] - r["dx"],10)==0
    
    r  = order_.selly(-500, raiseonerror=True)
    r
    
    assert r["y"] == 9500
    assert round(r["p"] - sqrt(r["pmarg_old"]*r["pmarg"]),10) == 0
    assert r["dx/dy"] == 1 / r["dy/dx"]
    assert r["tx"] == 'Buy 500 USDC sell ETH'
    assert r["expanded"] == False
    assert r["y_old"] - r["dy"] == r["y"]
    assert round(r["x"] - r["dx"],10)!=0
    

# ------------------------------------------------------------
# Test      044
# File      test_044_MarginalPriceRouting.py
# Segment   CarbonOrderUI charts [NOTEST]
# ------------------------------------------------------------
def notest_carbonorderui_charts():
# ------------------------------------------------------------
    
    order1 = CarbonOrderUI.from_prices("ETH/USDC", "ETH", 2000, 3000, 10, 10)
    order2 = CarbonOrderUI.from_prices("ETH/USDC", "ETH", 2000, 3000, 10, 5)
    orderr = CarbonOrderUI.from_prices("ETH/USDC", "USDC", 1000, 750, 25000, 25000)
    
    ETHr = np.linspace(0,order1.yint)
    USDCr = np.linspace(0,orderr.yint)
    
    plt.plot([order1.xfromy_f(y) for y in ETHr], ETHr, label="order1 (y=10)")
    plt.plot([order2.xfromy_f(y) for y in ETHr], ETHr, label="order2 (y=5)")
    plt.ylabel("y [ETH]")
    plt.xlabel("x [USDC]")
    plt.legend()
    
    plt.plot([order1.dxfromdy_f(y, raiseonerror=False) for y in ETHr], -ETHr, label="order1 (y=10)")
    plt.plot([order2.dxfromdy_f(y, raiseonerror=False) for y in ETHr], -ETHr, label="order2 (y=5)")
    plt.ylabel("dy [ETH]")
    plt.xlabel("dx [USDC]")
    plt.legend()
    
    plt.plot([order1.dxfromdy_f(y, raiseonerror=False) for y in ETHr], -ETHr, label="order1 (y=10)")
    plt.plot([order2.dxfromdy_f(y, raiseonerror=False) for y in ETHr], -ETHr, label="order2 (y=5)")
    plt.ylabel("dy [ETH]")
    plt.xlabel("dx [USDC]")
    plt.legend()
    
    plt.plot(
        [order1.dxfromdy_f(dy, raiseonerror=False) for dy in ETHr], 
        [order1.p_marg_f(dy, raiseonerror=False) for dy in ETHr], 
        label="marg (1; y=10 ETH)")
    plt.plot(
        [order2.dxfromdy_f(dy, raiseonerror=False) for dy in ETHr], 
        [order2.p_marg_f(dy, raiseonerror=False) for dy in ETHr], 
        label="marg (2; y=5 ETH)")
    plt.plot(
        USDCr, 
        [orderr.p_marg_f(dy, raiseonerror=False) for dy in USDCr], 
        label="marg(r; y=10k USDC)")
    plt.plot(
        [order1.dxfromdy_f(dy, raiseonerror=False) for dy in ETHr], 
        [order1.p_eff_f(dy, raiseonerror=False) for dy in ETHr], 
        label="eff (1; y=10 ETH)")
    plt.plot(
        [order2.dxfromdy_f(dy, raiseonerror=False) for dy in ETHr], 
        [order2.p_eff_f(dy, raiseonerror=False) for dy in ETHr], 
        label="eff (2; y=5 ETH)")
    plt.plot(
        USDCr, 
        [orderr.p_eff_f(dy, raiseonerror=False) for dy in USDCr], 
        label="eff (r; y=10k USDC)")
    plt.ylabel("Price [USDC per ETH]")
    plt.xlabel("dx [USDC]")
    plt.legend(loc="center right")
    

# ------------------------------------------------------------
# Test      044
# File      test_044_MarginalPriceRouting.py
# Segment   CarbonOrderUI tests goalseek
# ------------------------------------------------------------
def test_carbonorderui_tests_goalseek():
# ------------------------------------------------------------
    
    f = lambda x: sqrt(x)-3
    print (round(CarbonOrderUI.goalseek(f,1,10), 6))
    assert round(CarbonOrderUI.goalseek(f,1,10), 6) == 9
    
    try:
        CarbonOrderUI.goalseek(f,10,1)
        raise RuntimeError("Should raise exception")
    except ValueError as e:
        print(str(e))
    
    try:
        CarbonOrderUI.goalseek(f,10,20)
        raise RuntimeError("Should raise exception")
    except ValueError as e:
        print(str(e))
    

# ------------------------------------------------------------
# Test      044
# File      test_044_MarginalPriceRouting.py
# Segment   Example usage [NOTEST]
# ------------------------------------------------------------
def notest_example_usage():
# ------------------------------------------------------------
    #
    # This section shows how to use yfromp-related functions for routing. This section does not create test code.
    
    # First we create a staggered list of 10 orders. Note that all those orders are only half-filled. The corresponding marginal prices are shown in the printout.
    
    orders = [
        CarbonOrderUI.from_prices("ETH/USDC", "ETH", 2000+50*i, 2500+50*i, 10, 5)
        for i in range(10)
    ]
    for o in orders:
        print(f"y={o.y} pa={round(o.px,0)} pb={round(o.py,0)} pmarg={round(o.p_marg,1)}")
    margp = [o.p_marg for o in orders]
    min(margp), max(margp)
    
    # We are now looking at the dy released if we move to 2100. Spoiler: 0, because all positions are already beyond 2100.
    
    [o.dyfromp_f(2100) for o in orders]
    
    # If we go all the way to 3000 we see that all positions release their 5 ETH.
    
    [o.dyfromp_f(3000) for o in orders]
    
    # For 2300 we see that the first two positions can still get us to 2300. They release 1.39 and 0.40 ETH respectively on the way to 2300
    
    [o.dyfromp_f(2300) for o in orders]
    
    # For 2500, the first 6 positions release ETH
    
    [o.dyfromp_f(2500) for o in orders]
    
    # We define `dy_f` as the aggregate ETH released from all positions here
    
    dy_f = lambda p: sum(o.dyfromp_f(p) for o in orders)
    
    # The corresponding USDC inflow can be computed thus
    
    dx_f = lambda p: sum(o.dxfromdy_f(o.dyfromp_f(p)) for o in orders)
    
    # We can then plot the ETH release as a function of price
    
    pvals = np.linspace(2000, 3000)
    dyvals = [dy_f(p) for p in pvals]
    dxvals = [dx_f(p) for p in pvals]
    
    plt.plot(pvals, dyvals)
    plt.xlabel("Marginal price (USDC per ETH)")
    plt.ylabel("Aggregate ETH release")
    
    plt.plot(pvals, dxvals)
    plt.xlabel("Marginal price (USDC per ETH)")
    plt.ylabel("Aggregate USDC inflow")
    
    plt.plot(dxvals, dyvals)
    plt.xlabel("Aggregate USDC inflow")
    plt.ylabel("Aggregate ETH release")
    
    plt.plot(pvals, [dx/dy if dy>0 else None for dx,dy in zip(dxvals, dyvals)])
    plt.xlabel("Marginal price (USDC per ETH)")
    plt.ylabel("Effective price (USDC per ETH)")
    
    # Note for **Asaf**: this almost solves your "what is the price" problem if you plot it the other way round. The only issue is that here this is the marginal price after a move, so you'd have to integrate over it. We can also get the dx values the way we got the dy values (via the swap equation) but I do not have this implemented yet.
    
    plt.plot(dyvals, pvals)
    plt.ylabel("Marginal price (USDC per ETH)")
    plt.xlabel("Aggregate ETH release")
    
    # Now this does solve your price question
    
    plt.plot(dyvals, [dx/dy if dy>0 else None for dx,dy in zip(dxvals, dyvals)])
    plt.xlabel("Aggregate ETH release")
    plt.ylabel("Effective price (USDC per ETH)")
    
    # Going back to the routing problem we want to solve is that we have an aggregate release, say `15`, and we want the price; this we can do with goalseek
    
    p_goal = CarbonOrderUI.goalseek(lambda p: dy_f(p)-15, 2000, 3000)
    p_goal
    
    # The contribution of the positions at this price -- aka the **routing** -- is as follows
    
    rl1 = [o.dyfromp_f(p_goal) for o in orders]
    rl1
    
    # We can check that we routed indeed 15 ETH (the slight discrepancy is probably mostly related to goalseek precision)
    
    sum([o.dyfromp_f(p_goal) for o in orders])
    
    # And finally we can verify this against the exact algo match
    
    Sim = CarbonSimulatorUI(pair="ETH/USDC", verbose=False, raiseonerror=True)
    Sim
    
    for i in range(10):
        Sim.add_strategy('ETH', 10, 2000+50*i, 2500+50*i, None, None, None, None)
        # sell 5 ETH off each order to get the appropriate p_marg
        Sim.amm_sells('ETH', 5, use_positions=[i*2])
    Sim.state()['orders'].query("disabled==False")
    
    Sim.state()
    
    r = Sim.amm_sells('ETH', 15)['trades']
    r
    
    # Below `rl2` is the route list obtained through routing, and `rl1` the original one. We see they are extremely close.
    
    rl2 = list(r.query("aggr==False")["amt1"])
    rl2
    
    
    rl1[:len(rl2)]
    
    [x2-x1 for x1,x2 in zip(rl1, rl2)]
    
    


from carbon import CarbonSimulatorUI


def test_alpha_routing_trade_by_src():
    """
    Derived from notebook 4-1
    """
    AlphaSim = CarbonSimulatorUI(pair="ETH/USDC", verbose=False, matching_method=CarbonSimulatorUI.MATCH_ALPHA)
    strat = ['ETH', 10, 2000, 2500, 1000, 2800, 2700]
    for _ in range(5):
        AlphaSim.add_strategy(*strat)

    usdc_trade_amount = 1000
    eth_trade_amount = 0.5

    AlphaSim.amm_buys('USDC', usdc_trade_amount, execute=False, threshold_orders=3)  # route_trade_by_source
    assert AlphaSim.state()["trades"].iloc[3]["amt1"] == 0.499121771875

    AlphaSim.amm_buys('ETH', eth_trade_amount, execute=False, threshold_orders=3)  # route_trade_by_source
    assert AlphaSim.state()["trades"].query("uid=='1'").iloc[0]["amt1"] == 1388.325436499391


def test_alpha_routing_trade_by_target():
    """
    Derived from notebook 4-1
    """
    AlphaSim = CarbonSimulatorUI(pair="ETH/USDC", verbose=False, matching_method=CarbonSimulatorUI.MATCH_ALPHA)
    strat = ['ETH', 10, 2000, 2500, 1000, 2800, 2700]
    for _ in range(5):
        AlphaSim.add_strategy(*strat)

    usdc_trade_amount = 1000
    eth_trade_amount = 0.5

    AlphaSim.amm_sells('USDC', usdc_trade_amount, execute=False, threshold_orders=3)  # route_trade_by_target
    assert AlphaSim.state()["trades"].query("uid=='0'").iloc[0]["amt2"] == 0.35930099786

    AlphaSim.amm_sells('ETH', eth_trade_amount, execute=False, threshold_orders=2)  # route_trade_by_target
    assert AlphaSim.state()["trades"].query("uid=='1'").iloc[0]["amt2"] == 1002.64630467044

    AlphaSim.amm_sells('ETH', eth_trade_amount, execute=False, threshold_orders=4)  # route_trade_by_target
    assert AlphaSim.state()["trades"].query("uid=='2'").iloc[0]["amt2"] == 1001.321403916542



# ------------------------------------------------------------
# Auto generated test file `test_046_ThresholdOrders.py`
# ------------------------------------------------------------
# source file   = NBTest_046_ThresholdOrders.py
# source path   = /Users/skl/REPOES/Bancor/CarbonSimulator/resources/NBTest/
# target path   = /Users/skl/REPOES/Bancor/CarbonSimulator/resources/NBTest/
# test id       = 046
# test comment  = ThresholdOrders
# ------------------------------------------------------------



from carbon import CarbonSimulatorUI, __version__, __date__
print(f"Carbon Version v{__version__} ({__date__})", )





# ------------------------------------------------------------
# Test      046
# File      test_046_ThresholdOrders.py
# Segment   ThresholdOrdersUsePos
# ------------------------------------------------------------
def test_thresholdordersusepos():
# ------------------------------------------------------------
    
    Sim = CarbonSimulatorUI(pair="ETH/USDC", verbose=False)
    AlphaSim = CarbonSimulatorUI(pair="ETH/USDC", verbose=False, matching_method=CarbonSimulatorUI.MATCH_ALPHA)
    
    # +
    strat = ['ETH', 10, 2000, 2500, 1000, 2800, 2700]
    for _ in range(10):
        Sim.add_strategy(*strat)
        AlphaSim.add_strategy(*strat)
    
    for _ in range(3):
        Sim.add_strategy('ETH', 0.1, 1950, 1950, 1, 2750, 2750)
        AlphaSim.add_strategy('ETH', 0.1, 1950, 1950, 1, 2750, 2750)
    
    #Sim.state()['orders']
    AlphaSim.state()['orders']
    # -
    
    # set test amounts
    usdc_trade_amount = 1000
    eth_trade_amount = 0.5
    
    # ### Route by Source
    
    # #### AMM buys USDC
    
    threshold_orders=4
    use_positions=[0, 22, 24]
    
    Sim.amm_buys('USDC',usdc_trade_amount, execute=False)['trades']  # route_trade_by_source
    
    r = AlphaSim.amm_buys('USDC',usdc_trade_amount, execute=False, threshold_orders=threshold_orders)['trades']  # route_trade_by_source
    # threshold orders is only utilized on the alpha router
    # check that the maximum number of orders used is less than or equal to the number of threshold orders
    assert(len(r.query('subid!="A"')) <= threshold_orders)
    r
    
    r = Sim.amm_buys('USDC',usdc_trade_amount, execute=False, use_positions=use_positions)['trades']  # route_trade_by_source
    # use_positions is used on any router
    # check that the maximum number of orders used is less than or equal to the number of use_positions
    assert(len(r.query('subid!="A"')) <= len(use_positions))
    r
    
    r = AlphaSim.amm_buys('USDC',usdc_trade_amount, execute=False, threshold_orders=threshold_orders, use_positions=use_positions)['trades']  # route_trade_by_source
    # thus both use_positions and threshold should apply for alpha router
    assert(len(r.query('subid!="A"')) <= threshold_orders)
    assert(len(r.query('subid!="A"')) <= len(use_positions))
    r
    
    # #### AMM buys ETH
    
    threshold_orders=4
    use_positions=[1, 13, 19]
    
    Sim.amm_buys('ETH',eth_trade_amount, execute=False)['trades']  # route_trade_by_source
    
    r = AlphaSim.amm_buys('ETH',eth_trade_amount, execute=False, threshold_orders=threshold_orders)['trades']  # route_trade_by_source
    # threshold orders is only utilized on the alpha router
    # check that the maximum number of orders used is less than or equal to the number of threshold orders
    assert(len(r.query('subid!="A"')) <= threshold_orders)
    r
    
    r = Sim.amm_buys('ETH',eth_trade_amount, execute=False, use_positions=use_positions)['trades']  # route_trade_by_source
    # use_positions is used on any router
    # check that the maximum number of orders used is less than or equal to the number of use_positions
    assert(len(r.query('subid!="A"')) <= len(use_positions))
    r
    
    r = AlphaSim.amm_buys('ETH',eth_trade_amount, execute=False, threshold_orders=threshold_orders, use_positions=use_positions)['trades']  # route_trade_by_source
    # thus both use_positions and threshold should apply for alpha router
    assert(len(r.query('subid!="A"')) <= threshold_orders)
    assert(len(r.query('subid!="A"')) <= len(use_positions))
    r
    
    # ### Route by Target
    
    # #### AMM sells USDC
    
    # First we test the standard simulator which should route through all 5 positions
    
    Sim.amm_sells('USDC',usdc_trade_amount, execute=False)['trades']  # route_trade_by_target
    
    # Then we can set a threshold on the number of orders using the AlphaSim
    
    r = AlphaSim.amm_sells('USDC',usdc_trade_amount, execute=False, threshold_orders=threshold_orders)['trades']  # route_trade_by_source
    # threshold orders is only utilized on the alpha router
    # check that the maximum number of orders used is less than or equal to the number of threshold orders
    assert(len(r.query('subid!="A"')) <= threshold_orders)
    r
    
    r = Sim.amm_sells('USDC',usdc_trade_amount, execute=False, use_positions=use_positions)['trades']  # route_trade_by_source
    # use_positions is used on any router
    # check that the maximum number of orders used is less than or equal to the number of use_positions
    assert(len(r.query('subid!="A"')) <= len(use_positions))
    r
    
    r = AlphaSim.amm_sells('USDC',usdc_trade_amount, execute=False, threshold_orders=threshold_orders, use_positions=use_positions)['trades']  # route_trade_by_source
    # thus both use_positions and threshold should apply for alpha router
    assert(len(r.query('subid!="A"')) <= threshold_orders)
    assert(len(r.query('subid!="A"')) <= len(use_positions))
    r
    
    # #### AMM sells ETH
    
    threshold_orders=4
    use_positions=[0, 22, 24]
    
    Sim.amm_sells('ETH',eth_trade_amount, execute=False)['trades']  # route_trade_by_target
    
    r = AlphaSim.amm_sells('ETH',eth_trade_amount, execute=False, threshold_orders = threshold_orders)['trades'] # route_trade_by_target
    # threshold orders is only utilized on the alpha router
    # check that the maximum number of orders used is less than or equal to the number of threshold orders
    assert(len(r.query('subid!="A"')) <= threshold_orders)
    r
    
    r = Sim.amm_sells('ETH',eth_trade_amount, execute=False, use_positions=use_positions)['trades']  # route_trade_by_source
    # use_positions is used on any router
    # check that the maximum number of orders used is less than or equal to the number of use_positions
    assert(len(r.query('subid!="A"')) <= len(use_positions))
    r
    
    r = AlphaSim.amm_sells('ETH',eth_trade_amount, execute=False, threshold_orders=threshold_orders, use_positions=use_positions)['trades']  # route_trade_by_source
    # thus both use_positions and threshold should apply for alpha router
    assert(len(r.query('subid!="A"')) <= threshold_orders)
    assert(len(r.query('subid!="A"')) <= len(use_positions))
    r


# ------------------------------------------------------------
# Auto generated test file `test_043_SglOrder.py`
# ------------------------------------------------------------
# source file   = NBTest_043_SglOrder.py
# source path   = /Users/skl/REPOES/Bancor/CarbonSimulator/resources/NBTest/
# target path   = /Users/skl/REPOES/Bancor/CarbonSimulator/resources/NBTest/
# test id       = 043
# test comment  = SglOrder
# ------------------------------------------------------------



from carbon import CarbonSimulatorUI, P, __version__, __date__
print(f"Carbon v{__version__} ({__date__})")
print("{0.__name__} v{0.__VERSION__} ({0.__DATE__})".format(CarbonSimulatorUI))





# ------------------------------------------------------------
# Test      043
# File      test_043_SglOrder.py
# Segment   Single orders
# ------------------------------------------------------------
def test_single_orders():
# ------------------------------------------------------------
    
    # ### add_order as shortcut for add_strategy
    
    Sim = CarbonSimulatorUI(pair="ETH/USDC", raiseonerror=True, exclude_future=False)
    Sim
    
    # #### Base token (ETH)
    
    r = Sim.add_order("ETH", 10, 2000, 3000)["orders"]
    assert len(r) == 2
    r
    
    rr=r.query("tkn=='ETH'")
    assert rr.iloc[0]["y"] == 10
    assert rr.iloc[0]["disabled"] == False
    assert rr.iloc[0]["p_start"] == 2000
    ix = rr.iloc[0]["id"]
    lix = rr.iloc[0]["lid"]
    print(f"id={ix} lid={lix}")
    rr
    
    rr=r.query("tkn=='USDC'")
    assert rr.iloc[0]["y"] == 0
    assert not rr.iloc[0]["y"] is None
    assert rr.iloc[0]["disabled"] == True
    assert rr.iloc[0]["p_start"] is None
    ix2 = rr.iloc[0]["id"]
    lix2 = rr.iloc[0]["lid"]
    print(f"id={ix2} lid={lix2}")
    assert ix == lix2
    assert lix == ix2
    rr
    
    # #### Quote token (USDC)
    
    r = Sim.add_order("USDC", 1000, 1000, 750)["orders"]
    assert len(r) == 2
    r
    
    rr=r.query("tkn=='USDC'")
    assert rr.iloc[0]["y"] == 1000
    assert rr.iloc[0]["disabled"] == False
    assert rr.iloc[0]["p_start"] == 1000
    rr
    
    rr=r.query("tkn=='ETH'")
    assert rr.iloc[0]["y"] == 0
    assert not rr.iloc[0]["y"] is None
    assert rr.iloc[0]["disabled"] == True
    assert rr.iloc[0]["p_start"] is None
    rr
    
    # ### _add_replace_single_order
    #
    # this is a private function that allows to place single orders; this is an advanced feature and the API may change over time
    
    # #### Ensure it only works with future enabled
    
    try:
        SimNF = CarbonSimulatorUI(pair="ETH/USDC", raiseonerror=True)
        SimNF._add_replace_single_order("ETH", 20, 2000, 3000)
    except CarbonSimulatorUI.ExcludedFutureFunctionality as e:
        print(e)
        assert str(e) == "Feature disabled (us `exclude_future = False` to enable)"
    
    # #### Generating new orders (use with care; money disappears)
    
    r = Sim._add_replace_single_order("ETH", 20, 2000, 3000)
    ro=r["orders"]
    print(f"id={r['id']}, lid={r['lid']}")
    assert ro.iloc[0]["id"] == r['id']
    assert ro.iloc[0]["lid"] == r['lid']
    ro
    
    r = Sim._add_replace_single_order("USDC", None, 1000, 750)
    ro=r["orders"]
    print(f"id={r['id']}, lid={r['lid']}")
    assert ro.iloc[0]["id"] == r['id']
    assert ro.iloc[0]["lid"] == r['lid']
    ro
    
    r = Sim._add_replace_single_order("ETH", 20, None, None)
    ro=r["orders"]
    print(f"id={r['id']}, lid={r['lid']}")
    assert ro.iloc[0]["id"] == r['id']
    assert ro.iloc[0]["lid"] == r['lid']
    ro
    
    r = Sim._add_replace_single_order("USDC")
    ro=r["orders"]
    print(f"id={r['id']}, lid={r['lid']}")
    assert ro.iloc[0]["id"] == r['id']
    assert ro.iloc[0]["lid"] == r['lid']
    ro
    
    r = Sim.state()["orders"]
    assert len(r) == 8
    r
    
    # #### Replacing existing orders
    # note: this completely replaces those orders; there are no checks whether the curve and token are the same
    
    
    r = Sim._add_replace_single_order("ETH", 21, 2010, 3010, oid=4, lid=5)
    ro=r["orders"]
    ro
    
    r = Sim._add_replace_single_order("USDC", 1100, 1001, 751, oid=5, lid=4)
    ro=r["orders"]
    ro
    
    r = Sim.state()["orders"]
    assert len(r) == 8
    r
    
    rr = r.query("id==5 or id==4")
    assert len(rr) == 2
    assert list(rr["id"]) == [4,5]
    assert list(rr["lid"]) == [5,4]
    rr
    
    # #### Generating a collection account
    # We here link multiple positions to a single USD account; **note that this is _not_ functionality that the MVP would allow**
    
    Sim = CarbonSimulatorUI(pair="ETH/USDC", raiseonerror=True, exclude_future=False)
    Sim
    
    # Setting up the collection account in USDC
    
    r = Sim._add_replace_single_order("USDC")
    coid = r["id"]
    print(f"Collection order id coid={coid}")
    r["orders"]
    
    # Setting up an account selling ETH for USDC
    
    r = Sim._add_replace_single_order("ETH", 10, 2000, 2000, lid=coid)
    r["orders"]
    
    # Setting up an account selling WBTC for USDC
    
    r = Sim._add_replace_single_order("WBTC", 1, 10000, 10000, pair="WBTC/USDC", lid=coid)
    r["orders"]
    
    # Setting up an account selling BNT for USDC
    
    r = Sim._add_replace_single_order("BNT", 100000, 2, 2, pair="BNT/USDC", lid=coid)
    r["orders"]
    
    r = Sim.state()["orders"]
    assert len(r) == 4
    assert list(r["lid"]) == [coid]*4
    assert len(set(r["id"])) == 4
    r
    
    Sim.amm_sells("ETH", 5)
    r=Sim.state()["trades"]
    r
    
    r=Sim.state()["orders"]
    r0=r.query(f"id=={coid}")
    assert int(r0["y"]) == 5*2000
    r
    
    Sim.amm_sells("WBTC", 0.5, pair="WBTC/USDC")
    Sim.amm_sells("BNT", 50000, pair="BNT/USDC")
    r=Sim.state()["trades"]
    r
    
    r=Sim.state()["orders"]
    r0=r.query(f"id=={coid}")
    print(int(r0["y"]), 5*2000 + 0.5 * 10000 + 50000*2)
    assert int(r0["y"]) == int(5*2000 + 0.5 * 10000 + 50000*2)
    r
    
    


"""
represents a Carbon order book position

(c) Copyright Bprotocol foundation 2022.
Licensed under MIT
"""
__version__ = "1.0"
__date__ = "8/Dec/2022"

import pprint
from dataclasses import dataclass
from decimal import Decimal
from decimal import getcontext
from math import sqrt
from typing import NewType

from .pair import CarbonPair


getcontext().prec = 128

pp = pprint.PrettyPrinter(indent=4)

DecFloatInt = NewType("value", int or Decimal or float)


@dataclass
class Order:
    """
    Curve class to represent a curve on the market maker's order book
    """

    __VERSION__ = __version__
    __DATE__    = __date__

    # Production variables
    B: DecFloatInt = None       # = sqrt(price_low)
    S: DecFloatInt = None       # = sqrt(price_high) - sqrt(price_low)
    y_int: DecFloatInt = None   # the initial intercept of the curve

    # Original variables
    p_high: DecFloatInt = None  # = high price, also p_a, p_start
    p_low: DecFloatInt = None   # = low price, also p_b, p_low
    p_marginal: DecFloatInt = None  # = marginal price, also p_c, p_mid
    disabled: bool = False      # if True, prices used to be None
    _y: DecFloatInt = None      # the current balance of the curve

    # Alternate variables
    D: DecFloatInt = None       # the current intercept of the curve
    _C: DecFloatInt = None      # the current balance of the curve

    pair: CarbonPair = None
    tkn: str = None
    pair_name: str = None
    adj_n_mode: bool = False
    auto_convert_variables: bool = True
    id: int = 0
    linked_to_id: int = 0
    _reverseq: bool = False
    verbose: bool = False

    @property
    def y(self) -> Decimal:
        """
        The current balance of the curve... Equal to y_int upon genesis.
        """
        return self.y_int if self._y is None else self._y

    @y.setter
    def y(self, value: Decimal) -> None:
        """
        Setter for the current balance of the curve...
        """
        self._y = value

    @property
    def C(self) -> Decimal:
        """
        The current balance of the curve... Equal to D upon genesis.
        """
        return self.D if self._C == 0 else self._C

    @C.setter
    def C(self, value: Decimal) -> None:
        """
        Setter for the current balance of the curve...
        """
        self._C = value

    @property
    def p(self) -> Decimal:
        p_high: Decimal = self.p_high
        p_low: Decimal = self.p_low
        return (p_high * p_low).sqrt()

    @property
    def x0(self) -> Decimal:
        n: Decimal = self.n
        p: Decimal = self.p
        y_int: Decimal = max(self.y_int, Decimal("1"))
        return (y_int * (1 - n)) / (p * (2 - n))

    @property
    def k(self) -> Decimal:
        x0 = self.x0
        y0 = self.y0
        return x0 * y0

    @property
    def a(self):
        return self.k * (self.n - 1) - self.n * self.x0 * self.y

    @property
    def x(self) -> Decimal:
        x0: Decimal = self.x0
        n: Decimal = self.n
        y: Decimal = self.y
        y0: Decimal = self.y0
        return x0 * (n * y - y + y0 * (2 - n)) / (n * y + y0 * (1 - n))

    @property
    def y0(self) -> Decimal:
        y_int: Decimal = max(self.y_int, Decimal("1"))
        n: Decimal = self.n
        return (y_int * (1 - n)) / (2 - n)

    @property
    def n(self) -> Decimal:
        p_high: Decimal = self.p_high
        p_low: Decimal = self.p_low
        return (
            1 - (p_low / p_high) ** Decimal(".25")
            if not self.adj_n_mode
            else Decimal("0.000000000000000001")
        )

    @property
    def reverseq(self):
        return self._reverseq

    @reverseq.setter
    def reverseq(self, value):
        self._reverseq = value

    # Use the dataclass post-init method to initialize all variable sets
    def __post_init__(self):

        if self.p_high is not None and self.p_low is not None:
            self.p_high = Decimal(self.p_high)
            self.p_low = Decimal(self.p_low)

        if self.y_int is not None:
            self.y_int = Decimal(self.y_int)

        if self.p_marginal is not None:
            self.p_marginal = Decimal(self.p_marginal)

        if self.auto_convert_variables:

            # Init original variables if alternate variables are set
            if self.B is not None and self.S is not None:
                self.B = Decimal(self.B)
                self.S = Decimal(self.S)
                self.pb_raw = self.B * self.B
                self.pa_raw = (self.S + self.B) ** 2
                self.pa = 1 / self.pa_raw if self.reverseq else self.pa_raw
                self.pb = 1 / self.pb_raw if self.reverseq else self.pb_raw
                if self.pa < self.pb:
                    self.p_low = self.pa
                    self.p_high = self.pb
                else:
                    self.p_low = self.pb
                    self.p_high = self.pa

            # Init alternate variables if original variables are set
            elif self.p_low is not None and self.p_high is not None:

                if self.p_high < self.p_low:
                    paa = self.p_high
                    self.p_high = self.p_low
                    self.p_low = paa

                self.B = sqrt(self.p_low)
                self.S = sqrt(self.p_high) - sqrt(self.p_low)

            if self.y_int is not None:
                self.D = self.C = self._y = self.y_int

            elif self.y_int is None and self.p_marginal is not None:
                if type(self.p_marginal) is Decimal:
                    self.y_int = self._y * (self.p_high.sqrt() - self.p_low.sqrt()) / (self.p_marginal.sqrt() - self.p_low.sqrt())
                else:
                    self.y_int = float(self._y) * (sqrt(self.p_high) - sqrt(self.p_low)) / (sqrt(self.p_marginal) - sqrt(self.p_low))

            elif self.D is not None:
                self.y_int = self.C = self._y = self.D

            if self.p_high == self.p_low:
                self.adj_n_mode = True





"""
FLS - File Load Save (simple wrappers for loading and saving data)

:fsave:     save data into a file
:fload:     load data from a file
:join:      convenience wrapper for os.path.join

:VERSION HISTORY:

- v1.0: fload, fsave, join
- v1.0.1: minor change in output
- v1.1: json and yaml
- v1.2: copyright notice, license & canonic URL

:copyright:     (c) Copyright Stefan LOESCH / topaze.blue 2022; ALL RIGHTS RESERVED
:license:       [MIT](https://opensource.org/licenses/MIT)
:canonicurl:    https://github.com/topazeblue/TopazePublishing/blob/main/code/fls.py
"""
__VERSION__ = "1.2"
__DATE__ = "06/Jan/2023"

import os as _os
import gzip as _gzip
import json as _json
import yaml as _yaml

#########################################################
# FSAVE
def fsave(data, fn, path=None, binary=False, json=False, yaml=False, wrapper=None, quiet=False, compressed=False):
    """
    saves data to the file fn

    :data:          the data to be saved
    :fn:            the filename 
    :path:          the file path (default is "")
    :binary:        if True, the data is to be written as binary data*
    :json:          if True, the data is to be written as json data*
    :yaml:          if True, the data is to be written as yaml data*
    :wrapper:       a wrapper string where `{}` is replaced with the data
    :quiet:         if True, do not print info message
    :compressed:    use gz compression (implies binary*)

    *binary, json, yaml == True are mutually exclusive; behaviour is undefined otherwise
    """
    if path is None:
        path = "."
    ffn = _os.path.join(path, fn)

    if not quiet: print (f"[fsave] Writing {fn} to {path}")
    
    if compressed: binary = True
    ftype = "wb" if binary else "w"

    if json:
        data = _json.dumps(data)
    
    if yaml:
        data = _yaml.safe_dump(data)

    if not wrapper is None:
        data = wrapper.format(data)
    
    if compressed:
        data = _gzip.compress(data.encode())

    with open (ffn, ftype) as f:
        f.write(data)

#########################################################
# FLOAD
def fload(fn, path=None, binary=False, json=False, yaml=False, wrapper=None, quiet=False, compressed=False):
    """
    loads data from the file fn

    :fn:            the filename 
    :path:          the file path (default is "")
    :binary:        if the data is to be written as binary data
    :json:          if True, the data is to be written as json data*
    :yaml:          if True, the data is to be written as yaml data*
    :wrapper:       a wrapper string where `{}` is replaced with the data
    :quiet:         if True, do not print info message
    :compressed:    use gz compression (implies binary)

    *binary, json, yaml == True are mutually exclusive; behaviour is undefined otherwise
    """
    if path is None:
        path = "."
    ffn = _os.path.join(path, fn)

    if not quiet: print (f"[fload] Reading {fn} from {path}")

    if compressed: binary = True
    ftype = "rb" if binary else "r"
    with open (ffn, ftype) as f:
        data = f.read()
    
    if compressed:
        data = _gzip.decompress(data)

    if json:
        data = _json.loads(data)
    
    if yaml:
        data = _yaml.safe_load(data)

    if not wrapper is None:
        data = wrapper.format(data)
    return data

CSSWRAPPER = "\n<style>\n{}\n</style>\n"

#########################################################
# JOIN
def join(*args):
    "convenience wrapper for os.path.join"
    return _os.path.join(*args)


"""
Carbon helper module - read time series from data frame
"""
__VERSION__ = "1.0"
__DATE__ = "21/01/2023"

import pandas as pd

# those imports are for re-export
from os.path import join as j
import os

def dfread(fn):
    """reads dataframe from file and asserts format"""
    
    if fn[-7:] == ".pickle":
        df = pd.read_pickle(fn)
    else:
        df = pd.read_csv(fn)
    assert df.columns[0] == "time"
    assert df.columns[1] == "datetime"
    return df
    
def pdread(fn, datacol=None, indexcol=None):
    """
    reads a dataframe and returns a single column with index
    
    :fn:        the (full) filename
    :datacol:   name or index of the data col; None returns frame
    :indexcol:  name of the index col (default: "datetime")
    :returns:   pandas series
    """
    if indexcol is None: indexcol = "datetime"

    df = dfread(fn)
    df = df.set_index(indexcol)
    if datacol is None:
        return df.iloc[:, 1:]
    elif isinstance(datacol, str):
        return df[datacol]
    elif isinstance(datacol, int):
        return df.iloc[:, datacol+1]
    else:
        raise ValueError("datacol must be None, str or int", datacol)

def pdcols(fn):
    """
    reads a dataframe and returns a single column with index
    
    :fn:        the (full) filename
    :returns:   the column names (excluding the first two)
    """
    return dfread(fn).columns[2:]

def pathtime(path):
    """returns the time (in years) covered by the series `path`"""
    return (path.index[-1]-path.index[0])/pd.Timedelta(days=1)/365.25


# ------------------------------------------------------------
# Auto generated test file `test_038_CarbonPair.py`
# ------------------------------------------------------------
# source file   = NBTest_038_CarbonPair.py
# source path   = /Users/skl/REPOES/Bancor/CarbonSimulator/resources/NBTest/
# target path   = /Users/skl/REPOES/Bancor/CarbonSimulator/resources/NBTest/
# test id       = 038
# test comment  = CarbonPair
# ------------------------------------------------------------



from carbon import CarbonSimulatorUI, CarbonPair, P, __version__, __date__
print(f"Carbon v{__version__} ({__date__})")
print("{0.__name__} v{0.__VERSION__} ({0.__DATE__})".format(CarbonSimulatorUI))




#

# ------------------------------------------------------------
# Test      038
# File      test_038_CarbonPair.py
# Segment   CarbonPair
# ------------------------------------------------------------
def test_carbonpair():
# ------------------------------------------------------------
    #
    # Testing the new CarbonPair interface -- **BREAKING CHANGE**
    
    # +
    #help(CarbonPair)
    # -
    
    assert(P==CarbonPair)
    assert(str(P("ETH/USDC"))=="P('ETH/USDC')")
    
    str(P("ETH/USDC"))
    
    try:
        p = CarbonPair()
    except ValueError as e:
        print(e)
        assert(str(e)=="('If pair is None must provide tknb, tknq', None, None, None)")
    
    try:
        p = CarbonPair("ETH", "USDC")
    except ValueError as e:
        print(e)
        assert(str(e)=="""("Parameters are pair, tknb, tknq; did you mean `tknb='ETH', tknq='USDC'` ?", 'ETH', 'USDC', None)""")
    
    p = CarbonPair("", "ETH", "USDC")
    print(p)
    assert(str(p) == "P('ETH/USDC')")
    
    p = CarbonPair(tknb="ETH", tknq="USDC")
    print(p)
    assert(str(p) == "P('ETH/USDC')")
    
    p = CarbonPair(tknq="USDC", tknb="ETH")
    print(p)
    assert(str(p) == "P('ETH/USDC')")
    
    try:
        p = CarbonPair("ETHUSDC")
    except ValueError as e:
        print(e)
        assert(str(e)=="""('Illegal slashpair', 'ETHUSDC')""")
    
    p = CarbonPair("ETH/USDC")
    print(p)
    assert(str(p) == "P('ETH/USDC')")
    
    p = CarbonPair.from_isopair_and_tkn("ETHUSDC", "USDC")
    print(p)
    assert(str(p) == "P('ETH/USDC')")
    
    p = CarbonPair.from_isopair_and_tkn("ETHUSDC", "ETH")
    print(p)
    assert(str(p) == "P('ETH/USDC')")
    
    try:
        p = CarbonPair.from_isopair_and_tkn("ETHUSDC", "WBTC")
    except ValueError as e:
        print(e)
        assert(str(e) == "('Invalid token specification (tkn not part of isopair)', 'ETHUSDC', 'WBTC')")
    
    p = CarbonPair.create("ETH/USDC")
    print(p)
    assert(str(p) == "P('ETH/USDC')")
    
    p = CarbonPair.create("ETH", "USDC")
    print(p)
    assert(str(p) == "P('ETH/USDC')")
    
    pp = CarbonPair.create(p)
    print(pp)
    assert(str(pp) == "P('ETH/USDC')")
    assert(pp is p)
    
    try:
        pp = CarbonPair.create(p, "ETH")
    except ValueError as e:
        print(e)
        assert(str(e) == "('Second argument must be None if arg1 is pair', P('ETH/USDC'), 'ETH')")
    
    assert(p.tknb=="ETH")
    assert(p.tknq=="USDC")
    assert(p.basetoken=="ETH")
    assert(p.quotetoken=="USDC")
    assert(p.slashpair=="ETH/USDC")
    assert(p.pair_slash=="ETH/USDC")
    assert(p.pair_iso=="ETHUSDC")
    assert(p.price_convention=="USDC per ETH")
    assert(str(p.reverse)=="P('USDC/ETH')")
    
    assert(p.has_token("ETH"))
    assert(p.has_token("USDC"))
    assert(not p.has_token("WBTC"))
    
    assert(not p.has_quotetoken("ETH"))
    assert(p.has_quotetoken("USDC"))
    assert(not p.has_quotetoken("WBTC"))
    
    assert(p.has_basetoken("ETH"))
    assert(not p.has_basetoken("USDC"))
    assert(not p.has_basetoken("WBTC"))
    
    assert(p.other("ETH")=="USDC")
    assert(p.other("USDC")=="ETH")
    assert(p.other("WBTC") is None)
    
    # Convert an ETH amount into a USDC amount with the price in the price convention of the pair
    
    p.convert(1, "ETH", "USDC", 2000)
    
    # ditto ETH -> ETH (trivial)
    
    p.convert(1, "ETH", "ETH", 2000)
    
    # ditto USDC -> ETH (inverse rate!)
    
    p.convert(2000, "USDC", "ETH", 2000)
    
    assert(p.convert(1, "ETH", "USDC", 2000) == 2000)
    assert(p.convert(1, "ETH", "ETH", 2000) == 1)
    assert(p.convert(2000, "USDC", "ETH", 2000) == 1)
    assert(p.convert(1, "USDC", "USDC", 2000) == 1)
    
    assert(p.convert_price(2000, "USDC")==2000)
    assert(p.convert_price(1/2000, "ETH")==2000)
    
    # +
    #help(p.limit_is_met)
    # -
    
    p.limit_is_met("ETH", 2000, p.BUY, 1000, asphrase=True)
    
    p.limit_is_met("ETH", 2000, p.SELL, 1000, asphrase=True)
    
    p.limit_is_met("USDC", 2000, p.BUY, 1000, asphrase=True)
    
    p.limit_is_met("USDC", 2000, p.SELL, 1000, asphrase=True)
    
    assert(p.limit_is_met("ETH", 2000, p.BUY, 1000))
    assert(not p.limit_is_met("ETH", 2000, p.SELL, 1000))
    assert(not p.limit_is_met("USDC", 2000, p.BUY, 1000))
    assert(p.limit_is_met("USDC", 2000, p.SELL, 1000))
    
    
    


from decimal import Decimal
from decimal import getcontext
from decimal import ROUND_HALF_DOWN

getcontext().prec = 50
getcontext().rounding = ROUND_HALF_DOWN

def assertAlmostEqual(actual, expected, maxError):
    actual, expected, maxError = [Decimal(x) for x in [actual, expected, maxError]]
    if actual != expected:
        error = abs(actual - expected) / expected
        assert error <= maxError, 'error = {:f}'.format(error)



# ------------------------------------------------------------
# Auto generated test file `test_045_OrderBook.py`
# ------------------------------------------------------------
# source file   = NBTest_045_OrderBook.py
# source path   = /Users/skl/REPOES/Bancor/CarbonSimulator/resources/NBTest/
# target path   = /Users/skl/REPOES/Bancor/CarbonSimulator/resources/NBTest/
# test id       = 045
# test comment  = OrderBook
# ------------------------------------------------------------



from carbon import CarbonSimulatorUI, CarbonOrderUI, P, analytics as al, __version__, __date__
from math import sqrt
import numpy as np
from matplotlib import pyplot as plt
print(f"Carbon v{__version__} ({__date__})")
print("{0.__name__} v{0.__VERSION__} ({0.__DATE__})".format(CarbonSimulatorUI))
print("{0.__name__} v{0.__VERSION__} ({0.__DATE__})".format(CarbonOrderUI))




NUM_POINTS = 100   # number of points on the precise chart

ETHUSDC = P(tknq="USDC", tknb="ETH")
Sim = CarbonSimulatorUI(pair=ETHUSDC, verbose=False, raiseonerror=True)
orders = tuple([
    al.orders_nt("ETH", 100, 2000, 3000),
    al.orders_nt("ETH", 100, 2100, 2550),
    al.orders_nt("ETH", 50, 2300, 2450),
    al.orders_nt("ETH", 75, 2400, 2500),
    al.orders_nt("ETH", 80, 2500, 2700),
    al.orders_nt("USDC", 1000*150, 1500, 500),
    al.orders_nt("USDC", 1000*50, 1500, 1300),
    al.orders_nt("USDC", 1000*20, 1450, 1350),
    al.orders_nt("USDC", 1100*150, 1200, 1000),
])
for o in orders:
    Sim.add_order(o.tkn, o.amt, o.p_start, o.p_end)
Sim.add_order("WBTC", 1, 10000, 15000, pair="WBTC/USDC")
Sim.add_order("USDC", 1, 9000, 7500, pair="WBTC/USDC")
Sim.add_order("USDC", 10000, 0.99, 1.01, pair="USDC/USDT")
Sim.add_order("USDT", 10000, 1.01, 0.99, pair="USDC/USDT")
Sim.state()["orders"]

prices = al.linspace(500,3000, NUM_POINTS)
prices

curves_by_pair_bidask = CarbonOrderUI.curves_by_pair_bidask(Sim.state()["orderuis"])
print(list(curves_by_pair_bidask.keys()))


# ------------------------------------------------------------
# Test      045
# File      test_045_OrderBook.py
# Segment   curves_by_pair_bidask
# ------------------------------------------------------------
def test_curves_by_pair_bidask():
# ------------------------------------------------------------
    
    cbpba = curves_by_pair_bidask
    #cbpba["ETH/USDC"]
    
    print(list(cbpba.keys()))
    assert set(cbpba.keys()) == {'ETH/USDC', 'WBTC/USDC', 'USDC/USDT'}
    assert "ALL" in cbpba["WBTC/USDC"].keys()
    assert "BID" in cbpba["WBTC/USDC"].keys()
    assert "ASK" in cbpba["WBTC/USDC"].keys()
    assert len(cbpba["WBTC/USDC"]["ALL"])==len(cbpba["WBTC/USDC"]["BID"])+len(cbpba["WBTC/USDC"]["ASK"])
    assert len(cbpba["ETH/USDC"]["ALL"])==len(cbpba["ETH/USDC"]["BID"])+len(cbpba["ETH/USDC"]["ASK"])
    
    assert set(c.tkn for c in cbpba["ETH/USDC"]["ALL"]) == {'ETH', 'USDC'}
    assert set(c.tkn for c in cbpba["ETH/USDC"]["BID"]) == {'USDC'}
    assert set(c.tkn for c in cbpba["ETH/USDC"]["ASK"]) == {'ETH'}
    
    assert [int(c.p_marg) for c in cbpba["ETH/USDC"]["BID"] if not c.p_marg is None] == [1500, 1500, 1450, 1200]
    assert [int(c.p_marg) for c in cbpba["ETH/USDC"]["ASK"] if not c.p_marg is None] == [2000, 2100, 2300, 2399, 2500]
    

# ------------------------------------------------------------
# Test      045
# File      test_045_OrderBook.py
# Segment   CarbonOrderUI id and linked
# ------------------------------------------------------------
def test_carbonorderui_id_and_linked():
# ------------------------------------------------------------
    
    p1 = CarbonOrderUI.from_prices(ETHUSDC, "ETH", 2000, 3000, 10, 10)
    p2 = CarbonOrderUI.from_prices(ETHUSDC, "USDC", 1000, 750, 10, 10)
    p1,p2
    
    assert p1.id is None
    assert p1.linked is None
    
    p1.set_id(1)
    p2.set_id(2)
    assert p1.id == 1
    assert p2.id == 2
    
    p1.set_linked(p2)
    p2.set_linked(p1)
    assert p1.linked.id == 2
    assert p2.linked.id == 1
    
    p1.linked
    
    try:
        p1.set_id(1)
        raise RuntimeError("Should raise")
    except ValueError as e:
        print(e)
    
    try:
        p1.set_linked(p2)
        raise RuntimeError("Should raise")
    except ValueError as e:
        print(e)
    
    curves = curves_by_pair_bidask["ETH/USDC"]
    assert [c.id for c in curves["ALL"]] == [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17]
    assert [c.id for c in curves["BID"]] == [1, 3, 5, 7, 9, 10, 12, 14, 16]
    assert [c.id for c in curves["ASK"]] == [0, 2, 4, 6, 8, 11, 13, 15, 17]
    
    assert [c.linked.id for c in curves["ALL"]] == [1, 0, 3, 2, 5, 4, 7, 6, 9, 8, 11, 10, 13, 12, 15, 14, 17, 16]
    assert [c.linked.id for c in curves["BID"]] == [c.id for c in curves["ASK"]]
    assert [c.linked.id for c in curves["ASK"]] == [c.id for c in curves["BID"]]
    

# ------------------------------------------------------------
# Test      045
# File      test_045_OrderBook.py
# Segment   Approximate liquidity [NOTEST]
# ------------------------------------------------------------
def notest_approximate_liquidity():
# ------------------------------------------------------------
    
    liq =  al.calc_liquidity_approx(Sim.state()["orderuis"], prices, ETHUSDC, reverse=False)
    liqr = al.calc_liquidity_approx(Sim.state()["orderuis"], prices, ETHUSDC, reverse=True)
    
    al.plot_approx_orderbook_chart(liq)
    
    al.plot_approx_orderbook_chart(liqr)
    

# ------------------------------------------------------------
# Test      045
# File      test_045_OrderBook.py
# Segment   ASK _ AMM SELLS base token [NOTEST]
# ------------------------------------------------------------
def notest_ask___amm_sells_base_token():
# ------------------------------------------------------------
    
    curves = curves_by_pair_bidask["ETH/USDC"]["ASK"]
    c0 = curves[0]
    print(f"pair={c0.pair.slashpair} [{c0.pair.price_convention}] tkny={c0.tkny} tknx={c0.tknx}")
    dy_p = lambda p: sum(c.dyfromp_f(p) for c in curves)
    dx_p = lambda p: sum(c.dxfromdy_f(c.dyfromp_f(p)) for c in curves)
    dy_amounts = [dy_p(p) for p in prices]
    dx_amounts = [dx_p(p) for p in prices]
    
    plt.plot(prices, dy_amounts)
    plt.xlabel(f"market price [{c0.pair.price_convention}]")
    plt.ylabel(f"Cumulative amount of y sold [{c0.tkny}]")
    
    plt.plot(prices, dx_amounts)
    plt.xlabel(f"market price [{c0.pair.price_convention}]")
    plt.ylabel(f"Cumulative amount of x bought [{c0.tknx}]")
    
    OB = al.OrderBook(dy_amounts, dx_amounts, "ETH", "USDC")
    print(OB.explain())
    
    OB.plot_token_amount_chart()
    
    # When SELLING ETH, the AMM sells more and more expensively the more ETH it sells
    
    OB.plot_price_chart()
    
    # When SELLING ETH, the AMM pays more (in ETH terms) for the first units of USD received than for the later ones
    
    OB.plot_orderbook_chart()
    

# ------------------------------------------------------------
# Test      045
# File      test_045_OrderBook.py
# Segment   BID _ AMM BUYS base token [NOTEST]
# ------------------------------------------------------------
def notest_bid___amm_buys_base_token():
# ------------------------------------------------------------
    
    curves = curves_by_pair_bidask["ETH/USDC"]["BID"]
    c0 = curves[0]
    print(f"pair={c0.pair.slashpair} [{c0.pair.price_convention}] tkny={c0.tkny} tknx={c0.tknx}")
    dy_p = lambda p: sum(c.dyfromp_f(p) for c in curves)
    dx_p = lambda p: sum(c.dxfromdy_f(c.dyfromp_f(p)) for c in curves)
    dy_amounts = [dy_p(p) for p in prices]
    dx_amounts = [dx_p(p) for p in prices]
    
    OB2 = al.OrderBook(dx_amounts, dy_amounts, "ETH", "USDC", bidask=al.OrderBook.BID)
    print(OB2.explain())
    
    OB2.plot_token_amount_chart()
    
    OB2.plot_price_chart()
    
    OB2.plot_orderbook_chart()
    

# ------------------------------------------------------------
# Test      045
# File      test_045_OrderBook.py
# Segment   Combined [NOTEST]
# ------------------------------------------------------------
def notest_combined():
# ------------------------------------------------------------
    
    # Note: the liquidity is USDC liquidity, not ETH liquidity. This makes the numbers somewhat harder to verify. However -- this means that liquidity can be compared across different pairs that are using the same quote token.
    
    OB.plot_orderbook_chart(otherob=OB2)
    
    al.plot_approx_orderbook_chart(liqr)
    
    al.plot_approx_orderbook_chart(liq)
    
    


import subprocess
import sys

import pytest

from carbon import CarbonPair, CarbonSimulatorUI, analytics as al


def test_orderbook_explain():
    """
    Derived from `passed_tests` notebook demo-3-2
    """
    Sim = CarbonSimulatorUI(
        pair=CarbonPair("ETH/USDC"), verbose=False, raiseonerror=True
    )
    CA = al.Analytics(Sim, verbose=True)
    maxx = 3000
    orders = tuple(
        [
            al.orders_nt("ETH", 100, 2000, maxx),
            al.orders_nt("ETH", 100, 2500, 2700),
            al.orders_nt("USDC", 1000 * 100, 1500, 500),
            al.orders_nt("USDC", 1100 * 100, 1200, 1000),
        ]
    )

    for o in orders:
        Sim.add_order(o.tkn, o.amt, o.p_start, o.p_end)

    max_liquidity = Sim.liquidity()["ETHUSDC"]["ETH"]
    src_amounts = al.linspace(max_liquidity, 20)
    CA.simulate_trades(60, CA.ASK)
    CA.simulate_trades(70, CA.ASK)
    trg_amounts = al.vec([CA.simulate_trades(size, CA.ASK) for size in src_amounts])

    OB = al.OrderBook(src_amounts, trg_amounts, "ETH", "USDC")
    assert (
            OB.explain()
            == "This is the ASK book.\nSource token = ETH, target token = USDC.\nAMM sells ETH for USDC.\nBase token = ETH, quote token = USDC.\nPrices are quoted in USDC per ETH.\nOrder book amounts are quoted in USDC."
    )


def test_orderbook_plot_token_amount_chart_text():
    """
    Derived from `passed_tests` notebook demo-3-2
    """
    Sim = CarbonSimulatorUI(
        pair=CarbonPair("ETH/USDC"), verbose=False, raiseonerror=True
    )
    CA = al.Analytics(Sim, verbose=True)
    maxx = 3000
    orders = tuple(
        [
            al.orders_nt("ETH", 100, 2000, maxx),
            al.orders_nt("ETH", 100, 2500, 2700),
            al.orders_nt("USDC", 1000 * 100, 1500, 500),
            al.orders_nt("USDC", 1100 * 100, 1200, 1000),
        ]
    )

    for o in orders:
        Sim.add_order(o.tkn, o.amt, o.p_start, o.p_end)

    max_liquidity = Sim.liquidity()["ETHUSDC"]["ETH"]
    src_amounts = al.linspace0(max_liquidity, 20)
    CA.simulate_trades(60, CA.ASK)
    CA.simulate_trades(70, CA.ASK)
    trg_amounts = al.vec([CA.simulate_trades(size, CA.ASK) for size in src_amounts])

    OB = al.OrderBook(src_amounts, trg_amounts, "ETH", "USDC")
    result = OB.plot_tokenamount_chart()
    assert result == "plotted tokens received against trade size (504,750)"


# The following tests the matplotlib functionality using the `mpl` fixture
# https://github.com/matplotlib/pytest-mpl/blob/main/tests/test_pytest_mpl.py
#
# @pytest.mark.mpl_image_compare(baseline_dir='/carbon/tests/ui_tests/analytics/baseline',
#                                filename='orderbook_plot_token_amount_chart_1.png',
#                                tolerance=100)
# def test_orderbook_plot_token_amount_chart():
#     """
#     Derived from `passed_tests` notebook demo-3-2
#     """
#     Sim = CarbonSimulatorUI(pair=CarbonPair("ETH/USDC"), verbose=False, raiseonerror=True)
#     CA = al.Analytics(Sim, verbose=True)
#     maxx = 3000
#     orders = tuple([
#         al.orders_nt("ETH", 100, 2000, maxx),
#         al.orders_nt("ETH", 100, 2500, 2700),
#         al.orders_nt("USDC", 1000 * 100, 1500, 500),
#         al.orders_nt("USDC", 1100 * 100, 1200, 1000),
#     ])
#
#     for o in orders:
#         Sim.add_order(o.tkn, o.amt, o.p_start, o.p_end)
#
#     max_liquidity = Sim.liquidity()["ETHUSDC"]["ETH"]
#     src_amounts = al.linspace(max_liquidity, 20)
#     CA.simulate_trades(60, CA.ASK)
#     CA.simulate_trades(70, CA.ASK)
#     trg_amounts = al.vec([
#         CA.simulate_trades(size, CA.ASK) for size in src_amounts
#     ])
#
#     OB = al.OrderBook(src_amounts, trg_amounts, "ETH", "USDC")
#
#     return OB.plot_token_amount_chart(return_fig=True)



# ------------------------------------------------------------
# Auto generated test file `test_041_DisabledStrategies.py`
# ------------------------------------------------------------
# source file   = NBTest_041_DisabledStrategies.py
# source path   = /Users/skl/REPOES/Bancor/CarbonSimulator/resources/NBTest/
# target path   = /Users/skl/REPOES/Bancor/CarbonSimulator/resources/NBTest/
# test id       = 041
# test comment  = DisabledStrategies
# ------------------------------------------------------------



from carbon import CarbonSimulatorUI, P, __version__, __date__
print(f"Carbon v{__version__} ({__date__})")
print("{0.__name__} v{0.__VERSION__} ({0.__DATE__})".format(CarbonSimulatorUI))




Sim = CarbonSimulatorUI(pair="ETH/USDC", raiseonerror=True)
Sim


# ------------------------------------------------------------
# Test      041
# File      test_041_DisabledStrategies.py
# Segment   Disabled Orders
# ------------------------------------------------------------
def test_disabled_orders():
# ------------------------------------------------------------
    # ### Base token
    
    o = Sim.add_order("ETH", 100, 2000, 3000)["orders"]
    assert(o.iloc[0]["y"] == 100)
    assert(o.iloc[0]["y_int"] == 100)
    assert(o.iloc[0]["p_start"] == 2000)
    assert(o.iloc[0]["p_end"] == 3000)
    assert(o.iloc[0]["disabled"] == False)
    assert(o.iloc[0]["p_unit"] == "USDC per ETH")
    o
    
    o = Sim.add_order("ETH", None, 2000, 3000)["orders"]
    assert(o.iloc[0]["y"] == 0)
    assert(o.iloc[0]["y_int"] == 0)
    assert(o.iloc[0]["p_start"] == 2000)
    assert(o.iloc[0]["p_end"] == 3000)
    assert(o.iloc[0]["disabled"] == False)
    o
    
    o = Sim.add_order("ETH", None, 3000, 2000)["orders"]
    assert(o.iloc[0]["p_start"] == 2000)
    assert(o.iloc[0]["p_end"] == 3000)
    o
    
    o = Sim.add_order("ETH", 1, None, None)["orders"]
    assert(o.iloc[0]["y"] == 1)
    assert(o.iloc[0]["y_int"] == 1)
    assert(o.iloc[0]["disabled"] == True)
    assert(o.iloc[0]["p_start"] is None)
    assert(o.iloc[0]["p_end"] is None)
    assert(o.iloc[0]["p_marg"] is None)
    o
    
    o = Sim.add_order("ETH", None, None, None)["orders"]
    assert(o.iloc[0]["y"] == 0)
    assert(o.iloc[0]["y_int"] == 0)
    o
    
    # ### Quote token
    
    o = Sim.add_order("USDC", 1000, 1000, 750)["orders"]
    assert(o.iloc[0]["y"] == 1000)
    assert(o.iloc[0]["y_int"] == 1000)
    assert(o.iloc[0]["p_start"] == 1000)
    assert(o.iloc[0]["p_end"] == 750)
    assert(o.iloc[0]["disabled"] == False)
    assert(o.iloc[0]["p_unit"] == "USDC per ETH")
    o
    
    o = Sim.add_order("USDC", None, 1000, 750)["orders"]
    assert(o.iloc[0]["y"] == 0)
    assert(o.iloc[0]["y_int"] == 0)
    assert(o.iloc[0]["p_start"] == 1000)
    assert(o.iloc[0]["p_end"] == 750)
    assert(o.iloc[0]["disabled"] == False)
    o
    
    o = Sim.add_order("USDC", None, 750, 1000)["orders"]
    assert(o.iloc[0]["p_start"] == 1000)
    assert(o.iloc[0]["p_end"] == 750)
    o
    
    o = Sim.add_order("USDC", 1000, None, None)["orders"]
    assert(o.iloc[0]["y"] == 1000)
    assert(o.iloc[0]["y_int"] == 1000)
    assert(o.iloc[0]["disabled"] == True)
    assert(o.iloc[0]["p_start"] is None)
    assert(o.iloc[0]["p_end"] is None)
    assert(o.iloc[0]["p_marg"] is None)
    o
    
    o = Sim.add_order("USDC", None, None, None)["orders"]
    assert(o.iloc[0]["y"] == 0)
    assert(o.iloc[0]["y_int"] == 0)
    o
    
    
    # ### Fails
    
    try:
        Sim.add_order("USDC", 10, None, 1000)
    except ValueError as e:
        print(e)
        assert(str(e)=="('p_lo, p_hi must either be both None or both numbers', 1000, None)")
    
    try:
        Sim.add_order("ETH", 10, 2000, None)
    except ValueError as e:
        print(e)
        assert(str(e)=="('p_lo, p_hi must either be both None or both numbers', None, 0.0005)")
    

# ------------------------------------------------------------
# Test      041
# File      test_041_DisabledStrategies.py
# Segment   Disabled Strategies
# ------------------------------------------------------------
def test_disabled_strategies():
# ------------------------------------------------------------
    # ### Base token
    
    o = Sim.add_strategy("ETH", 10, 2000, 3000, 1000, 1000, 750)["orders"]
    assert(o.iloc[0]["disabled"] == False)
    assert(o.iloc[1]["disabled"] == False)
    assert(o.iloc[0]["p_start"] == 2000)
    assert(o.iloc[1]["p_start"] == 1000)
    assert(o.iloc[0]["p_end"] == 3000)
    assert(o.iloc[1]["p_end"] == 750)
    o
    
    o = Sim.add_strategy("ETH", 10, 3000, 2000, 1000, 750, 1000)["orders"]
    assert(o.iloc[0]["p_start"] == 2000)
    assert(o.iloc[1]["p_start"] == 1000)
    assert(o.iloc[0]["p_end"] == 3000)
    assert(o.iloc[1]["p_end"] == 750)
    o
    
    o = Sim.add_strategy("ETH", None, 2000, 3000, None, 1000, 750)["orders"]
    assert(o.iloc[0]["disabled"] == False)
    assert(o.iloc[1]["disabled"] == False)
    o
    
    o = Sim.add_strategy("ETH", 10, None, None, 1100, 1000, 750)["orders"]
    assert(o.iloc[0]["y"] == 10)
    assert(o.iloc[1]["y"] == 1100)
    assert(o.iloc[0]["disabled"] == True)
    assert(o.iloc[1]["disabled"] == False)
    o
    
    
    o = Sim.add_strategy("ETH", 10, 2000, 3000, 1100, None, None)["orders"]
    assert(o.iloc[0]["y"] == 10)
    assert(o.iloc[1]["y"] == 1100)
    assert(o.iloc[0]["disabled"] == False)
    assert(o.iloc[1]["disabled"] == True)
    o
    
    # ### Quote token
    
    o = Sim.add_strategy("USDC", 1000, 1000, 750, 10, 2000, 3000)["orders"]
    assert(o.iloc[0]["disabled"] == False)
    assert(o.iloc[1]["disabled"] == False)
    assert(o.iloc[1]["p_start"] == 2000)
    assert(o.iloc[0]["p_start"] == 1000)
    assert(o.iloc[1]["p_end"] == 3000)
    assert(o.iloc[0]["p_end"] == 750)
    o
    
    o = Sim.add_strategy("USDC", 1000, 750, 1000, 10, 3000, 2000)["orders"]
    assert(o.iloc[0]["disabled"] == False)
    assert(o.iloc[1]["disabled"] == False)
    assert(o.iloc[1]["p_start"] == 2000)
    assert(o.iloc[0]["p_start"] == 1000)
    assert(o.iloc[1]["p_end"] == 3000)
    assert(o.iloc[0]["p_end"] == 750)
    o
    
    o = Sim.add_strategy("USDC", 1100, None, None, 10, 2000, 3000)["orders"]
    assert(o.iloc[0]["disabled"] == True)
    assert(o.iloc[1]["disabled"] == False)
    assert(o.iloc[0]["y"] == 1100)
    assert(o.iloc[1]["y"] == 10)
    o
    
    o = Sim.add_strategy("USDC", 1100, None, None, 10, None, None)["orders"]
    assert(o.iloc[0]["disabled"] == True)
    assert(o.iloc[1]["disabled"] == True)
    assert(o.iloc[0]["y"] == 1100)
    assert(o.iloc[1]["y"] == 10)
    o
    

# ------------------------------------------------------------
# Test      041
# File      test_041_DisabledStrategies.py
# Segment   Trading
# ------------------------------------------------------------
def test_trading():
# ------------------------------------------------------------
    
    Sim = CarbonSimulatorUI(pair="ETH/USDC", raiseonerror=True)
    Sim.add_strategy("ETH", 10, 2000, 3000, 1000, 1000, 750)
    Sim.add_strategy("ETH", 10, None, None, 1000, 1000, 750)
    Sim.add_strategy("ETH", 10, 2000, 3000, 1000, None, None)
    Sim.add_strategy("ETH", None, 2000, 3000, None, 1000, 750)
    assert len(Sim.state()["orders"]) == 8
    Sim.state()["orders"]
    
    s = Sim.state()["orders"].query("disabled==False").query("y>0")
    assert list(s["id"]) == [0,1,3,4]
    s
    
    t = Sim.amm_sells("ETH", 10)["trades"]
    assert list(t["routeix"])[:-1] == [0,4]
    t
    
    s = Sim.state()["orders"].query("disabled==False").query("y>0")
    s
    
    t = Sim.amm_sells("USDC", 1000)["trades"]
    assert list(t["routeix"])[:-1] == [1,3]
    t
    
    


from carbon import CarbonSimulatorUI


def test_use_positions():
    """
    Test feature that allows the user to be able to provide a set of position numbers that the trade will be ONLY
    filled from those positions.
    """
    Sim = CarbonSimulatorUI(pair="ETH/USDC")
    Sim.add_strategy("ETH", 10, 2000, 2500, 10000, 1000, 750)
    Sim.add_strategy("ETH", 10, 2000, 2500, 10000, 1000, 750)
    Sim.add_strategy("ETH", 10, 2000, 2500, 10000, 1000, 750)
    Sim.add_strategy("ETH", 10, 2000, 2500, 10000, 1000, 750)
    Sim.add_strategy("ETH", 10, 2000, 2500, 10000, 1000, 750)
    Sim.add_strategy("ETH", 10, 2000, 2500, 10000, 1000, 750)
    result = Sim.trader_buys("ETH", 1, use_positions=[0, 2, 4])
    assert list(result['trades']['routeix'].values) == [0, 2, 4, '[0, 2, 4]']


def test_use_positions_with_missing():
    """
    Test that `use_positions` feature does not crash even if not both positions in a strategy are given.
    """
    Sim = CarbonSimulatorUI(pair="ETH/USDC")
    Sim.add_strategy("ETH", 10, 2000, 2500, 10000, 1000, 750)
    Sim.add_strategy("TEST", 10, 2000, 2500, 10000, 1000, 750, pair='TEST/USDC')
    result = Sim.trader_buys("ETH", 1, use_positions=[0, 3])
    assert list(result['trades']['routeix'].values) == [0, '[0]']


def test_use_positions_display():
    """
    Test that `use_positions` feature displays the proper routeidx.
    """
    Sim = CarbonSimulatorUI(verbose=True, raiseonerror=False, pair="ETH/USDC")
    Sim.add_order("ETH", 10, 2000, 2000)
    Sim.add_order("ETH", 10, 2010, 2010)
    Sim.add_order("ETH", 20, 2020, 2020)
    Sim.add_order("ETH", 30, 2030, 2030)
    Sim.add_order("ETH", 40, 2040, 2040)
    Sim.add_strategy("ETH", 10, 2050, 2050, 5000, 1060, 1060)
    Sim.add_strategy("ETH", 10, 2070, 2070, 5000, 1080, 1080)
    Sim.add_strategy("ETH", 10, 2090, 2090, 5000, 1100, 1100)
    result = Sim.amm_sells("ETH", 50, execute=False, use_positions=[3, 4])
    assert result['success'] == False


# ------------------------------------------------------------
# Auto generated test file `test_047_FastRouter.py`
# ------------------------------------------------------------
# source file   = NBTest_047_FastRouter.py
# source path   = /Users/skl/REPOES/Bancor/CarbonSimulator/resources/NBTest/
# target path   = /Users/skl/REPOES/Bancor/CarbonSimulator/resources/NBTest/
# test id       = 047
# test comment  = FastRouter
# ------------------------------------------------------------



import pandas as pd
from decimal import Decimal
from io import StringIO

from carbon import CarbonSimulatorUI, __version__, __date__
print(f"Carbon Version v{__version__} ({__date__})", )





# ------------------------------------------------------------
# Test      047
# File      test_047_FastRouter.py
# Segment   FastRouterTests
# ------------------------------------------------------------
def test_fastroutertests():
# ------------------------------------------------------------
    
    # Initialize a fast simulator
    
    Sim = CarbonSimulatorUI(pair="ETH/USDC", verbose=False, raiseonerror=True)
    FastSim = CarbonSimulatorUI(pair="ETH/USDC", verbose=False, matching_method='fast', raiseonerror=True)
    
    # Here we restrain the orders to just the top ones such that we can test the exact algo alongside
    
    test_orders = pd.DataFrame.from_dict(
        {'id': {0: '0',
      1: '1',
      2: '2',
      3: '3',
      6: '6',
      10: '10',
      246: '246',
      247: '247',
      248: '248',
      250: '250',
      251: '251'},
     'liquidity': {0: '254814732',
      1: '253827078',
      2: '252839424',
      3: '251851770',
      6: '248888808',
      10: '244938192',
      246: '245925846',
      247: '246913500',
      248: '247901154',
      250: '249876462',
      251: '250864116'},
     'lowest_rate': {0: '256',
      1: '255',
      2: '254',
      3: '253',
      6: '250',
      10: '246',
      246: '247',
      247: '248',
      248: '249',
      250: '251',
      251: '252'},
     'highest_rate': {0: '257',
      1: '256',
      2: '255',
      3: '254',
      6: '251',
      10: '247',
      246: '248',
      247: '249',
      248: '250',
      250: '252',
      251: '253'},
     'current_rate': {0: '257',
      1: '256',
      2: '255',
      3: '254',
      6: '251',
      10: '247',
      246: '248',
      247: '249',
      248: '250',
      250: '252',
      251: '253'}}
    )
    
    # +
    for i in test_orders.index:
        Sim.add_strategy('USDC', Decimal(test_orders.liquidity[i]), Decimal(test_orders.highest_rate[i]), Decimal(test_orders.lowest_rate[i]),  None, None, None)
        FastSim.add_strategy('USDC', Decimal(test_orders.liquidity[i]), Decimal(test_orders.highest_rate[i]), Decimal(test_orders.lowest_rate[i]),  None, None, None)
    
    FastSim.state()['orders']
    # -
    
    FastSim.state()['orders']
    
    # ### Route by Source
    
    # #### AMM buys ETH
    
    # Lets do some simple checks against the exact algo and then verify against the fast router results
    #
    # 1. Trade an amount against exact
    
    results = Sim.amm_buys('ETH',10000000, execute=False)['trades']  # route_trade_by_source
    assert(f"{results[results.uid=='0']['amt1'].values[0]:.0f}" == "2519667817")
    results
    
    # We can observe the price is 251.996 USDC per ETH.
    #
    # For trade_by_source, when we increase the trade amount we push further down the curve which means that effective price should get worse.
    
    results = Sim.amm_buys('ETH',10500000, execute=False)['trades']  # route_trade_by_source
    assert(f"{results[results.uid=='1']['amt1'].values[0]:.0f}" == "2643009195")
    results
    
    # And indeed for the exact algo, the price gets lower (251.996 vs 251.715), i.e. for every unit of ETH spent I get less units of USDC in return
    
    # We can then do a similar thing for the fast router
    
    results = FastSim.amm_buys('ETH',10000000, execute=False)['trades']  # route_trade_by_source
    assert(f"{results[results.uid=='0']['amt1'].values[0]:.0f}" == "2519667817")
    results
    
    # We see that the orders are filled the same and the corresponding price is as we saw before.
    #
    # We can then trade the higher amount and again, expect to see the effectice rate decrease.
    
    results = FastSim.amm_buys('ETH',10500000, execute=False)['trades']  # route_trade_by_source
    assert(f"{results[results.uid=='1']['amt1'].values[0]:.0f}" == "2643009195")
    results
    
    # Again, we see that the price has, as before, decreased appropriately.
    
    # ### Route by Target
    
    # #### AMM sells USDC
    
    # We can now test the route by target.
    #
    # First we do the counter trade - selling the output amount of USDC from the initial ETH trade
    
    results = Sim.amm_sells('USDC',2519667817, execute=False)['trades']  # route_trade_by_target
    assert(f"{results[results.uid=='2']['amt2'].values[0]:.0f}" == "10000000")
    results
    
    # And we see that the return amount is 10000000 we saw before - but importantly so is the price
    #
    # Now we expect that if we take less off the curve then the rate gets better
    
    results = Sim.amm_sells('USDC',1500000000, execute=False)['trades']  # route_trade_by_target
    assert(f"{results[results.uid=='3']['amt2'].values[0]:.0f}" == "5904869")
    results
    
    # Indeed we see that the price is better for a smaller amount is less (254.027 vs 251.996)
    
    # Now we do the same for the Fast router. Trade in the previous output amount
    
    results = FastSim.amm_sells('USDC',2519667817, execute=False)['trades']  # route_trade_by_target
    assert(f"{results[results.uid=='2']['amt2'].values[0]:.0f}" == "10000000")
    results
    
    # And importantly we see the same price to the exact algo
    
    # The test now is to see the price get better as the request amount is lowered
    
    results = FastSim.amm_sells('USDC',1500000000, execute=False)['trades']  # route_trade_by_target
    assert(f"{results[results.uid=='3']['amt2'].values[0]:.0f}" == "5904869")
    results
    
    # And indeed the rate gets better


# ------------------------------------------------------------
# Auto generated test file `test_039_Future.py`
# ------------------------------------------------------------
# source file   = NBTest_039_Future.py
# source path   = /Users/skl/REPOES/Bancor/CarbonSimulator/resources/NBTest/
# target path   = /Users/skl/REPOES/Bancor/CarbonSimulator/resources/NBTest/
# test id       = 039
# test comment  = Future
# ------------------------------------------------------------



from carbon import CarbonSimulatorUI, P, __version__, __date__
print(f"Carbon v{__version__} ({__date__})")
print("{0.__name__} v{0.__VERSION__} ({0.__DATE__})".format(CarbonSimulatorUI))





# ------------------------------------------------------------
# Test      039
# File      test_039_Future.py
# Segment   Future
# ------------------------------------------------------------
def test_future():
# ------------------------------------------------------------
    
    # +
    #help(CarbonSimulatorUI)
    # -
    
    Sim = CarbonSimulatorUI()
    SimXF = CarbonSimulatorUI(exclude_future=True)
    SimIF = CarbonSimulatorUI(exclude_future=False)
    
    assert(Sim.exclude_future==True)
    assert(SimXF.exclude_future==True)
    assert(SimIF.exclude_future==False)
    
    try:
        Sim._raise_if_future_restricted()
    except Sim.ExcludedFutureFunctionality as e:
        print(e)
        assert(str(e)=="Feature disabled (us `exclude_future = False` to enable)")
    
    try:
        SimXF._raise_if_future_restricted()
    except Sim.ExcludedFutureFunctionality as e:
        print(e)
        assert(str(e)=="Feature disabled (us `exclude_future = False` to enable)")
    
    assert(SimIF._raise_if_future_restricted()==None)
    
    


"""
represents a single, unidirectional carbon order and provides convenience methods for UI access

(c) Copyright Bprotocol foundation 2022. 
Licensed under MIT

VERSION HISTORY
- v1.3: order book helper functions (p_marg_f, yfromp_f, dyfromp_f, dyfromdx_f, dyfromdx_f, goalseek)
- v1.4: new methods: fromQxy, Q, Gamma, sellx, selly; yfromx_f, xfromy_f, p_eff_f, xint (1.3.1)
- v1.5: linked curves (beta); bidask, curves_by_pair_bidask, checks for B,S=0 (1.4.1)
- v1.6: linked curves incl trading (final), addliqy, tradeto; minor formula improvement (1.5.1)
- v1.6.1: bugfix
"""
__version__ = "1.6.1"
__date__ = "21/Jan/2023"

try:
    from .pair import CarbonPair
except:
    from pair import CarbonPair

from dataclasses import dataclass
from math import sqrt

@dataclass
class CarbonOrderUI:
    """
    current state of a single Carbon order, for UI purposes
    
    properties that are parameters of the constructor
    :pair:    the corresponding token pair (specifically, its CarbonPair record)
    :tkn:     the token that this position is selling
    :B:       the B-parameter; B = sqrt pb_raw
    :S:       the S-parameter; S = sqrt pa_raw - Sqrt pb_raw
    :yint:    the y-intercept of the curve (also its current maximum capacity)
    :y:       the current y-coordinate on the curve (also current token holdings)
    :id:      id of this curve*
    :linked:  the linked curve object*

    *Beta; interace may still change in respect to ix, lix and linked as well 
    as the associated functions
    
    other properties set by the constructor
    :pa_raw:   the pa parameter in native quotation (dy/dx)
    :pb_raw:   ditto pb
    :pa:       the pa paramter in the quotation appropriate for the pair
    :pb:       ditto pb
    :pmin:     the min of pa, pb (in the quotation appropriate for the pair)
    :pmax:     the max of pa, pb (in the quotation appropriate for the pair)
    
    NOTES
    - natively, prices are quoted in the convention dy/dx, where tkn is the quote asset;
    tkn is also the asset being sold, so the numeraire is always the asset being sold
    - pa, pb are read at the intercepts "left to right", so pa=py is is the y intercept price,
    and pb=px is the x intercept price 
    - the properties pa_raw and pb_raw correspond to the native pa, pb; the prperties pa, pb
    are quoted in the correct currency conventions
    - the properties py = pa and px = pb are aliases
    """
    __VERSION__ = __version__
    __DATE__    = __date__
    
    pair: CarbonPair
    tkn: str
    B: float
    S: float
    yint: float
    y: float
    id: any = None
    linked: any = None

    def __post_init__(self):
        self.pair = CarbonPair(self.pair)
        self.tkn = self.tkn.upper()
        if not self.pair.has_token(self.tkn):
            raise RuntimeError("token not part of pair", self.tkn, self.pair)
        self.pb_raw = self.B * self.B
        self.pa_raw = (self.S + self.B)**2
        if self.pa_raw == 0 and self.pb_raw == 0:
            self.pa_raw = None
            self.pa_raw = None
            self.disabled = True
        else:
            self.disabled = False
        self.reverseq = True if self.pair.has_basetoken(self.tkn) else False

        if not self.disabled:
            self.pa = 1./self.pa_raw if self.reverseq else self.pa_raw
            self.pb = 1./self.pb_raw if self.reverseq else self.pb_raw
            if self.pa < self.pb:
                self.pmin = self.pa
                self.pmax = self.pb
            else:
                self.pmin = self.pb
                self.pmax = self.pa  
        else:
            self.pmin = None
            self.pmax = None        
    
    def set_id(self, id):
        """sets curve index; raises if curve ID already set"""
        if not self.id is None:
            raise ValueError("Curve ID has already been set", id, self.id)
        self.id = id
    
    @property
    def lid(self):
        """returns ID of linked curve, or None"""
        if self.linked is None:
            return None
        return self.linked.id
        
    def set_linked(self, linked=None):
        """
        sets linked object and index

        :linked:    the linked curve (CarbonOrderUI object)
        """
        if not self.linked is None:
            raise ValueError("Linked object has already been set", linked.id, self.linked.id)
        self.linked = linked

    @classmethod
    def from_BSy(cls, pair, tkn, B, S, yint, y):
        """
        alternative* constructor, taking B,S,y

        :pair:    the corresponding token pair (specifically, its CarbonPair record)
        :tkn:     the token that this order is selling
        :B:       the B-parameter; B = sqrt pb_raw
        :S:       the S-parameter; S = sqrt pa_raw - Sqrt pb_raw
        :yint:    the y-intercept of the curve (also its current maximum capacity)
        :y:       the current y-coordinate on the curve (also current token holdings)

        *technically this is the same as the constructor, but use of `from_BSy` is recommended
        over use of the native class constructor in case of future implementation changes
        """
        return cls(
            pair=pair, 
            tkn=tkn, 
            B=B, 
            S= S, 
            yint=yint, 
            y=y
        )

    @classmethod
    def from_prices(cls, pair, tkn, pa, pb, yint, y):
        """
        alternative constructor, taking prices pa, pb and curve capacity yint
        
        :pair:    the corresponding token pair (specifically, its CarbonPair record)
        :tkn:     the token that this order is selling
        :pa:      the price at the y intercept, in quotation corresponding to the pair*
        :pb:      the price at the x intercept, in quotation corresponding to the pair*
        :yint:    the y-intercept of the curve (also its current maximum capacity)
        :y:       the current y-coordinate on the curve (also current token holdings)
        
        *in their native quotation, pa, pb = -dy/dx at the y-intercept and x-intercept
        respectively; as the function y(x) is convex we must have pa >= pb; as this can
        be confusing in reverse quotation we correct by exchanging pa, pb if pb < pa
        """
        pair = CarbonPair(pair)
        if yint<0:
            raise ValueError("yint must be non-negative", yint)
        if y>yint:
            raise ValueError("y must not be bigger than yint (y={y}, yint={yint})", yint, y)
        if y<0:
            raise ValueError("y must be non-negative", y)

        if pair.has_basetoken(tkn):
            pa = 1./pa
            pb = 1./pb
        
        if pa < pb:
            print("[from_prices] exchanging pa, pb")
            paa = pa
            pa = pb
            pb = paa
            
        B = sqrt(pb)
        S = sqrt(pa) - sqrt(pb)
        return cls(
            pair=pair, 
            tkn=tkn, 
            B=B, 
            S=S, 
            yint=yint, 
            y=y
        )
    
    @classmethod
    def from_Qxy(cls, pair, tkn, Q, xint, yint, y):
        """
        alternative constructor, taking convexity Q and curve capacities xint, yint
        """
        pair = CarbonPair(pair)
        if not pair.has_token(tkn):
            raise ValueError("token not part of the pair", tkn, pair)
        if xint<=0:
            raise ValueError("xint must be positive", xint)
        if yint<=0:
            raise ValueError("yint must be positive", yint)
        if y>yint:
            raise ValueError("y must not be bigger than yint (y={y}, yint={yint})", yint, y)
        if y<0:
            raise ValueError("y must be non-negative", y)

        p0 = yint/xint
        B = sqrt(p0*Q)
        S = sqrt(p0/Q) - B

        return cls(
            pair=pair, 
            tkn=tkn, 
            B=B, 
            S=S, 
            yint=yint, 
            y=y
        )

    @classmethod
    def from_order(cls, order):
        """
        alternative constructor, from an Order object

        :order:     the Order or CarbonOrderUI object
        """
        try: 
            yint = order.y_int
        except:
            yint = order.yint

        return cls(
            pair=order.pair, 
            tkn=order.tkn, 
            B=float(order.B), 
            S= float(order.S), 
            yint=float(yint), 
            y=float(order.y),
            id=order.id,
        )

    @property 
    def tkny(self):
        """the token on the y-axis (alias for tkn)"""
        return self.tkn
    
    @property 
    def tknx(self):
        """the token on the x-axis"""
        return self.pair.other(self.tkn)

    @property
    def bidask(self):
        """returns BID or ASK depending on what type curve it is"""
        if self.tkny == self.pair.basetoken:
            return "BID"
        else:
            return "ASK"
    
    @property
    def p0(self):
        """
        the average or effective price of the range, p0 = sqrt(pa, pb)
        """
        return sqrt(self.pa*self.pb)
    
    @property
    def px(self):
        """alias for pb"""
        return self.pb 
    p_end = px

    @property
    def py(self):
        """alias for pa"""
        return self.pa
    p_start = py
    
    @property
    def Q(self):
        """Q parameter = sqrt(Px/Py) < 1"""
        return sqrt(self.pb_raw/self.pa_raw)

    @property
    def Gamma(self):
        """Gamma parameter (also known as n) = 1 - sqrt(Q)"""
        return 1 - sqrt(self.Q)

    @property
    def widthr(self):
        """the width ratio of the range, widthr = pmax/pmin"""
        return self.pmax/self.pmin
    
    @property
    def widthpc(self):
        """the percentage width of the range, widthpc = (pmax-pmin)/p0"""
        return (self.pmax-self.pmin)/self.p0

    @property
    def price_convention(self):
        """
        the price convention of the prices quoted
        
        :raw:   if False (default) return convention for all prices except 
                pa_raw and pb_raw which is returned for raw=True
        """
        if raw:
            return self.pair.price_convention(self.reverseq)
        return self.pair.price_convention()

    def descr(self, full=False):
        """provides a description of the order and curve"""
        s1 = f"Sell {self.tkn} buy {self.pair.other(self.tkn)}"
        s2 = f"from {self.pa:.4f} to {self.pb:.4f} {self.price_convention()}"
        s2 = f"from {self.pa:.4f} to {self.pb:.4f} {self.price_convention}"
        if full:
            s3 = f" ({self.y} {self.tkn} on curve, {self.y/self.yint*100:.0f}% of capacity)"
        else:
            s3 = ""
        return f"{s1} {s2}{s3}"

    @property
    def p_marg(self):
        """
        the current marginal price of the range (alias for p_margf)
        """
        return self.p_marg_f(0)


    def p_marg_f(self, dy=0, fullcurve=False, checkbounds=True, raiseonerror=False):
        """
        the marginal price function of the range

        :dy:            the negative(!) change in liquidity value, expressed as 
                        positive number; y* = y - dy
        :fullcurve:     if True, ignore curve state y and use yint, ie y* = yint - dy
        :checkbounds:   if True (default), check that y is in the right range
        :raiseonerror:  if True, raises on all errors; otherwise (default) may return None
        :returns:       marginal price a y*, p(y*) in the price convention of the pair
        """
        #dydx = ((self.B * self.yint + self.S * self.y) / self.yint)**2
        if self.disabled:
            return None

        y = self.yint - dy if fullcurve else self.y - dy
        #print(f"[p_marg_f] dy={dy} y={y} s.y={self.y} s.yint={self.yint}", fullcurve)
        if checkbounds:
            if dy < 0:
                if raiseonerror:
                    raise ValueError("Trade size dy must be a non-negative number", dy)
                return None
            if y < 0:
                if raiseonerror:
                    raise ValueError("Trade size dy too big, results in y<0", dy, y, self.y)
                return None
        
        if self.yint == 0:
            if not self.y == 0 and dy == 0:
                raise ValueError("If yint=0 you must also have y,dy=0", yint, y, dy)
            dydx = ((self.B + self.S))**2
        else:
            dydx = ((self.B + self.S * y/self.yint))**2
        result = dydx if self.pair.has_quotetoken(self.tkn) else 1/dydx
        return result
    
    def p_eff_f(self, dy=0, fullcurve=False, checkbounds=True, raiseonerror=False):
        """
        the effective price function of the range

        :dy:            the negative(!) change in liquidity value, expressed as 
                        positive number; y* = y - dy
        :fullcurve:     if True, ignore curve state y and use yint, ie y* = yint - dy
        :checkbounds:   if True (default), check that y is in the right range
        :raiseonerror:  if True, raises on all errors; otherwise (default) may return None
        :returns:       effective price between y (or yint) and y* in the price convention of the pair
       """
        #dydx = ((self.B * self.yint + self.S * self.y) / self.yint)**2
        if self.disabled:
            return None

        y0 = self.yint if fullcurve else self.y
        y = y0 - dy
        if checkbounds:
            if dy < 0:
                if raiseonerror:
                    raise ValueError("Trade size dy must be a non-negative number", dy)
                return None
            if y < 0:
                if raiseonerror:
                    raise ValueError("Trade size dy too big, results in y<0", dy, y, self.y)
                return None

        if self.yint == 0:
            if not y0 == 0 and dy == 0:
                raise ValueError("If yint=0 you must also have y,dy=0", yint, y, dy)
            dydx = ((self.B + self.S))**2
        else:
            if dy==0:
                return self.p_marg_f(self.yint-y, checkbounds=False, fullcurve=True)
            dydx = dy / (self.xfromy_f(y) - self.xfromy_f(y0))
        
        result = dydx if self.pair.has_quotetoken(self.tkn) else 1/dydx
        return result
    
    class PriceOutOfBoundsError(ValueError): pass
    class PriceOutOfBoundsErrorBeyondStart(PriceOutOfBoundsError): pass
    class PriceOutOfBoundsErrorBeyondEnd(PriceOutOfBoundsError): pass
    class PriceOutOfBoundsErrorBeyondMarg(PriceOutOfBoundsError): pass

    def yfromp_f(self, p, checkbounds=True, raiseonerror=False):
        """
        returns y as a function of the target marginal price

        :p:             the target marginal price, in the price convention of the pair
        :checkbounds:   if True (default), check that y is in the right range
        :raiseonerror:  if True, raises upon error, else return None
        :returns:       the y value at which the marginal price is achieved
                        if beyond the end it returns 0, if beyond start or current y None
        """
        # dydx = ((B * yint + S * y) / yint)**2 = (B + S y/yint)**2
        # y = yint * (sqrt(dydx) - B) / S
        dydx = p if self.pair.has_quotetoken(self.tkn) else 1/p
        #print(f"[yfromp_f] pa={self.pa_raw} dydx={dydx} pb={self.pb_raw}")

        if checkbounds:
            if dydx > self.pa_raw:
                if raiseonerror:
                    raise self.PriceOutOfBoundsErrorBeyondStart("Price out of bounds (beyond start)", p, self.pa)
                return None
            elif dydx < self.pb_raw:
                if raiseonerror:
                    raise self.PriceOutOfBoundsErrorBeyondEnd("Price out of bounds (beyond end)", p, self.pb)
                return 0
        y = self.yint * (sqrt(dydx) - self.B) / self.S
        if checkbounds:
            if y > self.y:
                if raiseonerror:
                    raise self.PriceOutOfBoundsErrorBeyondMarg("Price out of bounds (beyond marginal), hence target y > y", y, self.y )
                return None
        return y

    @property
    def xint(self):
        """
        x intercept, ie max x liquidity, ie x at y=0
        """
        #return self.yint**2 / (self.B**2*self.yint + self.B*self.S*self.yint)
        return self.yint / (self.B**2 + self.B*self.S)
        
    def dyfromp_f(self, p, checkbounds=True, raiseonerror=False):
        """
        returns dy = y_target - y as a function of the target marginal price

        :p:             the target marginal price, in the price convention of the pair
        :checkbounds:   if True (default), check that y is in the right range
        :raiseonerror:  if True, raises upon error, else return None
        :returns:       the (positive!) dy value at which the marginal price is achieved
                        in cases where yfromp_f returns none, this func returns 0
        """
        if self.B == 0 and self.S == 0:
            if raiseonerror:
                raise ValueError("Can't determine trade prices from an empty curve", self.B, self.S)
            return 0
        y = self.yfromp_f(p, checkbounds, raiseonerror)
        if y is None: return 0
        return self.y-y

    def dyfromdx_f(self, dx, checkbounds=True, raiseonerror=False):
        """
        calculates the amount dy SOLD by the AMM to RECEIVE an amount dx

        :dx:            the amount of x the AMM RECEIVES (a POSITIVE number*)
        :checkbounds:   if True (default), check that dy is in the right range
        :raiseonerror:  if True, raises upon error, else return None
        :returns:       the amount dy of y the AMM SELLS (a POSITIVE number*)

        *when checkbounds is False then we can have dx<0, corresponding to the 
        AMM SELLing x. In this case it returns a negative number, corresponding
        to the AMM BUYing y.
        """
        if self.B == 0 and self.S == 0:
            if raiseonerror:
                raise ValueError("Can't trade on an empty curve", self.B, self.S)
            return 0

        if checkbounds:
            if dx < 0:
                if raiseonerror:
                    raise ValueError("AMM buy amount dx must be a non-negative number", dx)
                return None
            # elif dx > self.y:
            #     if raiseonerror:
            #         raise ValueError("AMM sell amount dx must be within available liquidity", dx, self.y)
            #     return None
        
        num   =               (self.S*self.y + self.B*self.yint)**2
        #         -------------------------------------------------------------
        denom =   self.S*dx * (self.S*self.y + self.B*self.yint) + self.yint**2

        if checkbounds:
            if num < 0:
                if raiseonerror:
                    raise ValueError("AMM does not have enough y liquidity to purchase dx", y, dx, num)
                return None

        return dx * (num/denom) 


    def dxfromdy_f(self, dy, checkbounds=True, raiseonerror=False):
        """
        calculates the amount dx RECEIVED for a trade of dy

        :dy:            the amount of y the AMM sells (a POSITIVE number*)
        :checkbounds:   if True (default), check that dy is in the right range
        :raiseonerror:  if True, raises upon error, else return None
        :returns:       the amount of x the AMM receives (a POSITIVE number*)

        *when checkbounds is False then we can have dy<0, corresponding to the 
        AMM BUYing y. In this case it returns a negative number, corresponding
        to the AMM SELLing x.
        """
        if self.B == 0 and self.S == 0:
            if raiseonerror:
                raise ValueError("Can't trade on an empty curve", self.B, self.S)
            return 0

        if checkbounds:
            if dy < 0:
                if raiseonerror:
                    raise ValueError("AMM sell amount dy must be a non-negative number", dy)
                return None
            elif dy > self.y:
                if raiseonerror:
                    raise ValueError("AMM sell amount dy must be within available liquidity", dy, self.y)
                return None
        
        num   =                                   self.yint**2
        #       ----------------------------------------------------------------------------------
        denom = (self.S*self.y+self.B*self.yint) * (self.S*self.y+self.B*self.yint-self.S*dy)

        return dy*(num/denom)     
    
    def xfromy_f(self, y, checkbounds=True, raiseonerror=False):
        """
        the invariant function, expressed as x=f(y)

        :y:             the amount of y the AMM currently holds
        :checkbounds:   if True (default), check that dy is in the right range
        :raiseonerror:  if True, raises upon error, else return None
        :returns:       the corresponding x amount*

        *in Carbon, the funds received on this curve are not kept on this curve but 
        transferred to a linked curve; the linked curve may hold different amounts
        of x, eg because it has been seeded with x>0. In other words -- it does not
        make sense to look at absolute values of x, only of differences.    
        """
        if checkbounds:
            if y > self.yint:
                if raiseonerror:
                    raise ValueError("The value of y is out of bounds (y>yint)", y, self.yint)
                return None
            elif y < 0:
                if raiseonerror:
                    raise ValueError("The value of y is out of bounds (y<0)", y)
                return None

        num =                          self.yint * (self.yint - y)
        #        -----------------------------------------------------------------------------
        denom =  self.B**2*self.yint + self.B*self.S*y + self.B*self.S*self.yint + self.S**2*y
        return num/denom

    @property
    def x(self):
        """
        the current x value of the curve (=xfromy_f(self.y))
        """
        return self.xfromy_f(self.y)
    

    def yfromx_f(self, x, checkbounds=True, raiseonerror=False):
        """
        the invariant function, expressed as y=f(x)

        :x:             the theoretical* amount of x the AMM currently holds
        :checkbounds:   if True (default), check that dy is in the right range
        :raiseonerror:  if True, raises upon error, else return None
        :returns:       the corresponding      

        *in Carbon, the funds received on this curve are not kept on this curve but 
        transferred to a linked curve; the linked curve may hold different amounts
        of x, eg because it has been seeded with x>0. In other words -- it does not
        make sense to look at absolute values of x, only of differences.
        """
        if checkbounds:
            if x < 0:
                if raiseonerror:
                    raise ValueError("The value of x is out of bounds (x<0)", x)
                return None
            # if x < self.xint:
            #     if raiseonerror:
            #         raise ValueError("The value of x is out of bounds (x>xint; y<0)", x, result)
            #     return None 

        num =    self.yint*(-self.B**2*x - self.B*self.S*x + self.yint)
        #        ------------------------------------------------------
        denom =        self.B*self.S*x + self.S**2*x + self.yint

        y = num/denom
        if checkbounds:
            if y < 0:
                if raiseonerror:
                    raise ValueError("The value of x is out of bounds (x>xint; y<0)", x, xint, y)
                return None 
        return y

    
    @property
    def total_liquidity(self):
        """
        returns the total liquidity of the position, and the token in which it is quoted

        :returns:       (liquidity, token)
        """
        return self.y, self.tkn
    
    def liquidity_approx(self, price1, price2, tkn=None, asperc=False):
        """
        returns the approximate liquidity between start and end, in tkn

        :price1/2:      the start and end price of the range (in any order; quoted in price convention of pair)
        :tkn:           the token in which the liquidity is quoted (if None: base token)
        :asperc:        if True, return percentage total liquidity rather than tkn number; default is False
        :returns:       the liquidity in [price1, price2], quoted in tkn (or percent)
        """

        # ensure that price1 <= price2
        if price1 > price2:
            pp = price2
            price2 = price1
            price1 = pp

        if tkn is None:
            tkn = self.pair.basetoken
            #print(f"[liquidity_approx] token set to {tkn}")
            
        # # now we need to go through all the different arrangements of the range
        # # we first compute the percentage coverage, which simply is the percentage
            
        # # price1>pmax -> completely above the range
        # if price1 > self.pmax:
        #     perc = 0.
        
        # # price2<pmin -> completely below the range
        # elif price2 < self.pmin:
        #     perc = 0.

        # # price1 <= pmin
        # elif price1 <= self.pmin:

        #     # price1 <= min and price2 >= max -> 100%
        #     if price2 >= self.pmax:
        #         perc = 1.
        #     else:
        #         perc = (price2 - self.pmin)/(self.pmax - self.pmin)

        # # price2 >= pmax and price1>pmin
        # elif price2 >= self.pmax:

        #     perc = (self.pmax - price1)/(self.pmax - self.pmin)

        # # price2 < pmax and price1 > pmin
        # else:

        #     perc = (price2 - price1)/(self.pmax - self.pmin)

        # self.pmax == self.pmin -> 100% in range, otherwise out
        if self.disabled:
            return 0
            
        if self.pmax == self.pmin:
            if price1 == price2:
                perc2 = 1. if price1 == self.pmax else 0
            else:
                perc2 = 1. if price1 <= self.pmax and price2 > self.pmax else 0
            price = self.pmax
            #print(f"perc2={perc2}, price={price}", price1, price2, self.pmin, price1==self.pmax, price2==self.pmax)

        # price1 > self.pmax or price2 < self.pmin -> completely above or below the range
        elif price1 >= self.pmax or price2 <= self.pmin:
            perc2 = 0

        # alternatively: restrict p1,p2 to the ranges and calculate the coverage
        else:
            if price1 < self.pmin: price1 = self.pmin
            if price2 > self.pmax: price2 = self.pmax
            perc2 = (price2 - price1)/(self.pmax - self.pmin)
            price = sqrt(price1*price2)
        
        if asperc:
            return perc2
        
        if perc2 == 0:
            return 0

        # liquidity
        liq0 = self.y
        liq0_tkn = self.tkn

        # convert price into the correct quantity
        liq = self.pair.convert(amtfrom=liq0, tknfrom=liq0_tkn, tknto=tkn, price=price)
        
        return liq*perc2
    
    @staticmethod
    def goalseek(func, a, b, eps=None):
        """
        helper method: solves for x, a<x<b, such that func(x) == 0
        
        :func:    a function f(x), eg lambda x: x-3
        :a:       the lower bound
        :b:       the upper bound
        :eps:     precision, ie b/a value where goal seek returns
        :returns: the x value found
        """
        if eps is None:
            eps = 0.0000001
        if not a<b:
            raise ValueError("Bracketing value a must be smaller than b", a, b)
        fa = func(a)
        fb = func(b)
        if not fa*fb<0:
            raise ValueError("Sign of f(a) must be opposite of sign of f(b)", fa, fb, a, b)
            
        while 1:
            m = 0.5*(a+b)
            fm = func(m)
            if fm * fa > 0:
                a = m
            else:
                b = m
            
            #print(f"m={m}, m={m}, b={b}")
            if b/a-1 < eps:
                return m

    def selly(self, dy, execute=True, allowneg=True, expandcurve=False, raiseonerror=False):
        """
        executes a trade selling dy for dx (dy given)

        :dy:            the amount of y to sell (a positive number)
        :execute:       if False, only display results but do not update the object
        :allowneg:      if True, negative dy numbers (=buying) are allowed
        :expandcurve:   if True, purchasing y beyond yint [sic] expands the curve to yint = y
                        only meaningful with allowneg = True
        :raiseonerror:  if True, error lead to raising on exception
        """
        if dy < 0:
            if not allowneg:
                if raiseonerror:
                    raise ValueError(f"Negative dy is not allowed (allowneg={allowneg})", dy)
                return None

        if dy == 0:
            execute = False

        yold = self.y
        pold = self.p_marg
        ynew = self.y - dy
        pnew = self.p_marg_f(dy, checkbounds=False)

        if ynew < 0:
            if raiseonerror:
                raise ValueError(f"Traded beyond capacity (yold={yold}, ynew={ynew}, dy={dy})")
            return None

        elif ynew > self.yint:
            if not expandcurve:
                if raiseonerror:
                    raise ValueError(f"Curve needs expanding and expanding not allowed (yold={yold}, ynew={ynew}, dy={dy}, expandcurve={expandcurve})")
                return None
            else:
                yintold = self.yint
                yint = ynew
                if execute:
                    self.yint = ynew
                curve_expanded = True
        
        else:
            curve_expanded = False
            yint = self.yint

        dx = self.dxfromdy_f(dy, checkbounds=False, raiseonerror=True)
        if execute:
            self.y = ynew

        if self.linked:
            result_linked = self.linked.addliqy(dx)
        else:
            result_linked = None

        result = {
            "action":       "bysource[selly]",
            "executed":     execute,
            "y_old":        yold,
            "y":            ynew,
            "dy":           dy,
            "yint_old":     yintold if curve_expanded else None,
            "y_int":        yint,
            "expanded":     curve_expanded,
            "x":            self.xfromy_f(ynew),
            "dx":           dx,
            "tkny":         self.tkny,
            "tknx":         self.tknx,
            "tx":           f"Sell {abs(dy)} {self.tkny} buy {self.tknx}" if dx>0 else f"Buy {abs(dy)} {self.tkny} sell {self.tknx}",
            "dx/dy":        dx/dy if dy != 0 else None,
            "dy/dx":        dy/dx if dx != 0 else None,
            "pmarg_old":    pold,
            "pmarg":        pnew,
            "linked":       result_linked,
            "p":            None,
        } 
        result["p"] = result["dx/dy"] if self.pair.has_basetoken(self.tkny) else result["dy/dx"]
        return result

    def buyx(self, dx, execute=True, allowneg=True, expandcurve=False, raiseonerror=False):
        """
        executes a trade dy for dx (dx given)

        :dx:            the amount of x to buy (a positive number)
        :execute:       if False, only display results but do not update the object
        :allowneg:      if True, negative dy numbers (=buying) are allowed
        :expandcurve:   if True, purchasing y beyond yint [sic] expands the curve to yint = y
                        only meaningful with allowneg = True
        :raiseonerror:  if True, error lead to raising on exception
        """
        dy = self.dyfromdx_f(dx, checkbounds=False, raiseonerror=True)
        result = self.selly(dy, execute, allowneg, expandcurve, raiseonerror)
        result["action"] = "bytarget[buyx]"
        return result

    def tradeto(self, p_marg, execute=True, raiseonerror=False):
        """
        executes a trade dy for dx (target marginal price p_marg given)

        :p:             the target marginal price
        :execute:       if False, only display results, but do not update the object
        :raiseonerror:  if True, error lead to raising on exception
        """
        if self.yint == 0:
            self.yint = 1e-50 # selly fails on a fully empty curve, even with dy=0
        try:
            dy = self.dyfromp_f(p_marg, checkbounds=True, raiseonerror=True)
        except (self.PriceOutOfBoundsErrorBeyondStart, 
                self.PriceOutOfBoundsErrorBeyondMarg):
            dy = 0
        except self.PriceOutOfBoundsErrorBeyondEnd:
            dy = self.dyfromp_f(self.p_end, checkbounds=False, raiseonerror=False)
        result = self.selly(dy, execute, allowneg=False, expandcurve=False, raiseonerror=raiseonerror)
        result["action"] = "byprice[tradeto]"
        return result
        
    def addliqy(self, dy, expandcurve=True):
        """
        adds liquidity to the curve (typically called on a linked curve)

        :dy:            the amount of liquidity to be added (must be positive)
        :expandcurve:   if True (default), expand yint=y if need be
        """
        if dy < 0:
            raise ValueError("Liquidity amount dy must not be < 0", dy)
        
        newy = self.y + dy
        result = {
            "y_old":        self.y,
            "y":            newy,
            "dy":           dy,
            "yint_old":     self.yint,
            "yint":         self.yint,
            "tkny":         self.tkn,
            "pmarg_old":    self.p_marg,
            "pmarg":        None,
            "expanded":     False,
        }
        if newy > self.yint:
            if not expandcurve:
                raise ValueError("Must expand curve and expandcurve=False", self.y, dy, newy, self.yint)
            self.yint = newy
            result["yint"] = newy
            result["expanded"] = True
        self.y = newy
        result["pmarg"] = self.p_marg
        return result
        
    @staticmethod
    def curves_by_pair_bidask(curves):
        """
        sorts curves by pair, and then by bid/ask

        :curves:        an iterable of CarbonOrderUI curves
        :returns:       dict {
                            "pair": {
                                "ALL": [curve, ...],
                                "BID": [curve, ...],
                                "ASK": [curve, ...]
                            }
                            ...
                        }
        """
        pairs = set(r.pair.slashpair for r in curves.values())
        result = {
            pair:{ 
                "ALL": [v for v in curves.values() if v.pair.slashpair == pair],
                "ASK": [v for v in curves.values() if v.pair.slashpair == pair and v.bidask=="BID"],
                "BID": [v for v in curves.values() if v.pair.slashpair == pair and v.bidask=="ASK"],
            }
            for pair in pairs
        }            
        return result
    
    def __repr__(self):
        s1 = f"pair={self.pair.slashpair}, tkn={self.tkn}, B={self.B}, S={self.S}, yint={self.yint}, y={self.y}, id={self.id}"
        s2 = f"linked=<{self.linked.id}>" if self.linked else "linked=None"
        return f"{self.__class__.__name__}({s1}, {s2})" 


# ------------------------------------------------------------
# Auto generated test file `test_049_LimitPrices.py`
# ------------------------------------------------------------
# source file   = NBTest_049_LimitPrices.py
# source path   = /Users/skl/REPOES/Bancor/CarbonSimulator/resources/NBTest/
# target path   = /Users/skl/REPOES/Bancor/CarbonSimulator/resources/NBTest/
# test id       = 049
# test comment  = LimitPrices
# ------------------------------------------------------------



from carbon import CarbonSimulatorUI, CarbonOrderUI, P, __version__, __date__
from math import sqrt
import numpy as np
from matplotlib import pyplot as plt
print(f"Carbon v{__version__} ({__date__})")
print("{0.__name__} v{0.__VERSION__} ({0.__DATE__})".format(CarbonSimulatorUI))
print("{0.__name__} v{0.__VERSION__} ({0.__DATE__})".format(CarbonOrderUI))





# ------------------------------------------------------------
# Test      049
# File      test_049_LimitPrices.py
# Segment   Limit prices
# ------------------------------------------------------------
def test_limit_prices():
# ------------------------------------------------------------
    
    SimR = CarbonSimulatorUI(pair="ETH/USDC", raiseonerror=True)
    SimNR = CarbonSimulatorUI(pair="ETH/USDC", raiseonerror=False)
    SimR, SimNR
    
    SimR.add_order("ETH", 10, 2000, 3000)
    SimNR.add_order("ETH", 10, 2000, 3000)
    SimR.add_order("USDC", 10000, 1000, 500)
    SimNR.add_order("USDC", 10000, 1000, 500)
    assert len(SimR.state()["orders"]) == 4
    assert len(SimNR.state()["orders"]) == 4
    
    help(SimR._trade)
    
    assert SimR.trader_buys == SimR.amm_sells
    assert SimR.trader_sells == SimR.amm_buys
    assert SimR.trader_buys != SimR.amm_buys
    assert SimR.trader_sells != SimR.amm_sells
    help(SimR.trader_buys)
    help(SimR.trader_sells)
    
    r = SimR.amm_sells("ETH", 1, execute=False)
    assert r["success"] == True
    r["trades"]
    
    # ### Check that you can't give both price and amount, both by amount and by target
    
    try:
        SimR.amm_sells("ETH", 1, limit_price=1000, limit_amt=1000)
        raise RuntimeError("should raise")
    except ValueError as e:
        print(e)
    
    try:
        SimR.amm_buys("ETH", 1, limit_price=1000, limit_amt=1000)
        raise RuntimeError("should raise")
    except ValueError as e:
        print(e)
    
    try:
        SimR.amm_sells("USDC", 1, limit_price=1000, limit_amt=1000)
        raise RuntimeError("should raise")
    except ValueError as e:
        print(e)
    
    try:
        SimR.amm_buys("USDC", 1, limit_price=1000, limit_amt=1000)
        raise RuntimeError("should raise")
    except ValueError as e:
        print(e)
    
    
    assert SimNR.amm_sells("ETH", 1, limit_price=1000, limit_amt=1000)["success"] == False
    assert SimNR.amm_buys("ETH", 1, limit_price=1000, limit_amt=1000)["success"] == False
    assert SimNR.amm_sells("USDC", 1, limit_price=1000, limit_amt=1000)["success"] == False
    assert SimNR.amm_buys("USDC", 1, limit_price=1000, limit_amt=1000)["success"] == False
    
    # ### Check price limits
    
    r = SimR.trader_buys("ETH", 1, execute=False)
    price_buy_eth = r["trades"]["price"].iloc[0]
    assert int(price_buy_eth) == 2037
    
    r = SimR.trader_buys("USDC", 1000, execute=False)
    price_buy_usdc = r["trades"]["price"].iloc[0]
    assert int(price_buy_usdc) == 970
    
    r = SimR.trader_sells("ETH", 1, execute=False)
    price_sell_eth = r["trades"]["price"].iloc[0]
    assert int(price_sell_eth) == 971
    
    r = SimR.trader_sells("USDC", 1000, execute=False)
    price_sell_usdc = r["trades"]["price"].iloc[0]
    assert int(price_sell_usdc) == 2018
    
    print(f"buy  ETH   @ {price_buy_eth:10.0f} USDC/ETH")
    print(f"buy  USDC  @ {price_buy_usdc:10.0f} USDC/ETH")
    print(f"sell ETH   @ {price_sell_eth:10.0f} USDC/ETH")
    print(f"sell USDC  @ {price_sell_usdc:10.0f} USDC/ETH")
    
    limitfail = lambda  r: "fail" if r["trades"].query("aggr==True")["limitfail"].iloc[0] else "nofail"
    
    # worse than limit transactions
    
    fail_r = (
        limitfail(SimR.trader_buys("ETH", 1, limit_price = int(price_buy_eth) - 1, execute=False)),
        limitfail(SimR.trader_buys("USDC", 1000, limit_price = int(price_buy_usdc) + 1, execute=False)),
        limitfail(SimR.trader_sells("ETH", 1, limit_price = int(price_sell_eth) + 1, execute=False)),
        limitfail(SimR.trader_sells("USDC", 1000, limit_price = int(price_sell_usdc) - 1, execute=False)),
    )
    assert fail_r == ("fail",)*4
    fail_r
    
    # better than limit transactions
    
    fail_r = (
        limitfail(SimR.trader_buys("ETH", 1, limit_price = int(price_buy_eth) + 1, execute=False)),
        limitfail(SimR.trader_buys("USDC", 1000, limit_price = int(price_buy_usdc) - 1, execute=False)),
        limitfail(SimR.trader_sells("ETH", 1, limit_price = int(price_sell_eth) - 1, execute=False)),
        limitfail(SimR.trader_sells("USDC", 1000, limit_price = int(price_sell_usdc) + 1, execute=False)),
    )
    assert fail_r == ("nofail",)*4
    fail_r
    
    # ### Check amount limits
    
    # worse than limit transactions
    
    fail_r = (
        limitfail(SimR.trader_buys("ETH", 1, limit_amt = (int(price_buy_eth) - 1)*1, execute=False)),
        limitfail(SimR.trader_buys("USDC", 1000, limit_amt = 1000/(int(price_buy_usdc) + 1), execute=False)),
        limitfail(SimR.trader_sells("ETH", 1, limit_amt = (int(price_sell_eth) + 5)*1, execute=False)),
        limitfail(SimR.trader_sells("USDC", 1000, limit_amt = 1000/(int(price_sell_usdc) - 1), execute=False)),
    )
    assert fail_r == ("fail",)*4
    fail_r
    
    price_sell_usdc
    
    price_sell_usdc
    
    # better than limit transactions
    
    fail_r = (
        limitfail(SimR.trader_buys("ETH", 1, limit_amt = (int(price_buy_eth) + 1)*1, execute=False)),
        limitfail(SimR.trader_buys("USDC", 1000, limit_amt = 1000/(int(price_buy_usdc) - 1), execute=False)),
        limitfail(SimR.trader_sells("ETH", 1, limit_amt = (int(price_sell_eth) - 1)*1, execute=False)),
        limitfail(SimR.trader_sells("USDC", 1000, limit_amt = 1000/(int(price_sell_usdc) + 1), execute=False)),
    )
    assert fail_r == ("nofail",)*4
    fail_r
    
    
    # ### Check amount limits (ranges)
    
    # +
    def goalseek(f, target, xlo, xhi):
        if not f(xhi)>target and f(xlo)<target:
            raise ValueError("must have f(xhi) > target, f(xlo)<target ", f(xhi), f(xlo))
        return _goalseek(f, target, xlo, xhi)
        
    def _goalseek(f, target, xlo, xhi):
        xmid = 0.5*(xhi+xlo)
        #print(xlo, xhi, xmid)
        if abs(xhi-xlo)<1: return xmid
        if f(xmid) > target:
            return _goalseek(f, target, xlo, xmid)
        else:
            return _goalseek(f, target, xmid, xhi)
    limitfail_ind = lambda  r: 0 if r["trades"].query("aggr==True")["limitfail"].iloc[0] else 1
    # -
    
    x = goalseek(lambda x: limitfail_ind(SimR.trader_buys("ETH", 1, limit_amt=x, execute=False)), 0.5, 500, 3000)
    assert int(x) == int(price_buy_eth)
    
    x = goalseek(lambda x: limitfail_ind(SimR.trader_sells("ETH", 1, limit_amt=x, execute=False)), 0.5, 3000, 500)
    assert int(x) == int(price_sell_eth)
    
    


from carbon import CarbonSimulatorUI


def test_unrelated_token_in_add_order():
    Sim = CarbonSimulatorUI(pair="ETH/USDC", verbose=False, raiseonerror=False)

    response = Sim.add_order("LINK", 100, 2000, 2000)
    assert response["success"] is False
    # assert (
    #     response["error"]
    #     == "('Invalid token specification (tkn not part of isopair)', 'ETHUSDC', 'LINK')"
    # )

    response = Sim.add_order("DNE", 100, 2000, 2000)
    assert response["success"] is False
    # assert (
    #     response["error"]
    #     == "('Invalid token specification (tkn not part of isopair)', 'ETHUSDC', 'DNE')"
    # )



import math

from carbon import CarbonSimulatorUI


def test_liquidity_approx_order0():
    """
    Test Order 0 (ETH, range) - Derived from Test36-LiquidityApprox.ipynb
    """
    Sim = CarbonSimulatorUI(verbose=False, raiseonerror=False, pair="ETH/USDC")
    Sim.add_order("ETH", 10, 2000, 3000)
    Sim.add_order("ETH", 10, 2500, 2500)
    Sim.add_order("USDC", 10*1250, 1500, 1000)
    Sim.add_order("USDC", 10*1250, 1250, 1250)

    oui = Sim.state()["orderuis"][0]
    assert (oui.pmin, oui.pmax, oui.total_liquidity) == (2000.0, 3000.0, (10.0, 'ETH'))
    assert oui.liquidity_approx(500, 600, "ETH", asperc=True) == 0.0
    assert oui.liquidity_approx(3000, 3020, "ETH", asperc=True) == 0.0
    assert oui.liquidity_approx(2000, 3000, "ETH", asperc=True) == 1.0
    assert oui.liquidity_approx(3000, 2000, "ETH", asperc=True) == 1.0
    assert oui.liquidity_approx(2000, 2500, "ETH", asperc=True) == 0.5
    assert oui.liquidity_approx(2500, 3000, "ETH", asperc=True) == 0.5
    assert oui.liquidity_approx(2250, 2750, "ETH", asperc=True) == 0.5
    assert oui.liquidity_approx(2000, 3000, "ETH", asperc=False) == 10.0
    assert oui.liquidity_approx(2000, 2500, "ETH", asperc=False) == 5.0
    assert oui.liquidity_approx(2000, 2500, asperc=False) == 5.0
    assert oui.liquidity_approx(2000, 3000, "USDC", asperc=False) == math.sqrt(2000*3000)*10
    assert oui.liquidity_approx(2250, 2750, "USDC", asperc=False) == math.sqrt(2250*2750)*0.5*10


def test_liquidity_approx_order1():
    """
    Test Order 1 (ETH, point) - Derived from Test36-LiquidityApprox.ipynb
    """
    Sim = CarbonSimulatorUI(verbose=False, raiseonerror=False, pair="ETH/USDC")
    Sim.add_order("ETH", 10, 2000, 3000)
    Sim.add_order("ETH", 10, 2500, 2500)
    Sim.add_order("USDC", 10*1250, 1500, 1000)
    Sim.add_order("USDC", 10*1250, 1250, 1250)

    oui = Sim.state()["orderuis"][1]
    assert oui.total_liquidity == (0.0, 'USDC')
    assert oui.liquidity_approx(2000, 2499, "ETH", asperc=True) == 0.0
    assert oui.liquidity_approx(2501, 3020, "ETH", asperc=True) == 0.0
    assert oui.liquidity_approx(2499, 2501, "ETH", asperc=True) == 0.0
    assert oui.liquidity_approx(2500, 2501, "ETH", asperc=True) == 0.0
    assert oui.liquidity_approx(2499, 2500, "ETH", asperc=True) == 0.0
    assert oui.liquidity_approx(2500, 2500, "ETH", asperc=True) == 0.0
    assert oui.liquidity_approx(2499, 2501, "ETH", asperc=False) == 0.0
    assert oui.liquidity_approx(2499, 2501, "USDC", asperc=False) == 0.0
    assert oui.liquidity_approx(2500, 2501, "USDC", asperc=False) == 0.0
    assert oui.liquidity_approx(2500, 2500, "USDC", asperc=False) == 0.0
    assert oui.liquidity_approx(2000, 3000, "USDC", asperc=False) == 0.0


def test_liquidity_approx_order2():
    """
    Test Order 2 (USDC, range) - Derived from Test36-LiquidityApprox.ipynb
    """
    Sim = CarbonSimulatorUI(verbose=False, raiseonerror=False, pair="ETH/USDC")
    Sim.add_order("ETH", 10, 2000, 3000)
    Sim.add_order("ETH", 10, 2500, 2500)
    Sim.add_order("USDC", 10*1250, 1500, 1000)
    Sim.add_order("USDC", 10*1250, 1250, 1250)


    oui = Sim.state()["orderuis"][2]
    assert (oui.pmin, oui.pmax, oui.total_liquidity) == (2500.0, 2500.0, (10.0, 'ETH'))
    assert oui.liquidity_approx(500, 600, "USDC", asperc=False) == 0.0
    assert oui.liquidity_approx(1600, 2000, "USDC", asperc=False) == 0.0
    assert oui.liquidity_approx(1000, 1500, "USDC", asperc=False) == 0.0
    assert oui.liquidity_approx(1000, 1100, "USDC", asperc=False) == 0.0
    assert oui.liquidity_approx(1400, 1500, "USDC", asperc=False) == 0.0
    assert round(oui.liquidity_approx(1000, 1100, "ETH", asperc=False),6) == 0
    assert round(oui.liquidity_approx(1400, 1500, "ETH", asperc=False),6) == 0


def test_liquidity_approx_order3():
    """
    Test Order 3 (USDC, point) - Derived from Test36-LiquidityApprox.ipynb
    """
    Sim = CarbonSimulatorUI(verbose=False, raiseonerror=False, pair="ETH/USDC")
    Sim.add_order("ETH", 10, 2000, 3000)
    Sim.add_order("ETH", 10, 2500, 2500)
    Sim.add_order("USDC", 10*1250, 1500, 1000)
    Sim.add_order("USDC", 10*1250, 1250, 1250)

    oui = Sim.state()["orderuis"][3]
    assert oui.total_liquidity == (0.0, 'USDC')
    assert oui.liquidity_approx(1249, 1251, "USDC", asperc=False) == 0.0
    assert oui.liquidity_approx(1000, 2000, "USDC", asperc=False) == 0.0
    assert oui.liquidity_approx(1249, 1250, "USDC", asperc=False) == 0.0
    assert oui.liquidity_approx(1250, 1251, "USDC", asperc=False) == 0.0
    assert oui.liquidity_approx(1249, 1251, "ETH", asperc=False) == 0.0
    assert oui.liquidity_approx(1000, 2000, "ETH", asperc=False) == 0.0
    assert oui.liquidity_approx(1249, 1250, "ETH", asperc=False) == 0.0
    assert oui.liquidity_approx(1250, 1251, "ETH", asperc=False) == 0.0